"""
TODO Nu Must Klein: Voeg laag toe aan GeoPackage met visuele koppeling tussen intrede en uittredepunten.
"""

from __future__ import annotations
from geopandas import GeoDataFrame, read_file
from pathlib import Path
from geoprob_pipe.cmd_app.spatial_joins.utils import append_to_gis_join_parameter_invoer_table
from geoprob_pipe.utils.validation_messages import BColors
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def coupled_distances_to_uittredepunten(app_settings: ApplicationSettings) -> bool:

    # Read uittredepunten
    gdf_exit_points: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="uittredepunten")

    # Check if already added
    columns = gdf_exit_points.columns
    if "afstand_intredelijn" in columns and "afstand_buitenteenlijn" in columns and "afstand_binnenteenlijn" in columns:
        print(BColors.OKBLUE,
              f"✔  Afstanden intrede, buitenteen en binnenteen al gekoppeld aan uittredepunten.", BColors.ENDC)
        return True  # Assuming already added
    # TODO: Uitfaseren dat deze kolommen gekoppeld worden aan de exit_points tabel

    # Distance to intredelijn
    gdf_intredelijnen: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="intredelijn")
    gdf_exit_points['afstand_intredelijn'] = gdf_exit_points.geometry.apply(
        lambda pnt: round(gdf_intredelijnen.distance(pnt).min(), 1))
    df_l_intrede = gdf_exit_points[["uittredepunt_id", "afstand_intredelijn"]]
    df_l_intrede = df_l_intrede.rename(columns={"afstand_intredelijn": "mean"})
    append_to_gis_join_parameter_invoer_table(
        df_sjoin=df_l_intrede, parameter_name="L_intrede", geopackage_filepath=app_settings.geopackage_filepath)

    # Distance to buitenteenlijn
    gdf_buitenteenlijnen: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="buitenteenlijn")
    gdf_exit_points['afstand_buitenteenlijn'] = gdf_exit_points.geometry.apply(
        lambda pnt: round(gdf_buitenteenlijnen.distance(pnt).min(), 1))
    df_l_but = gdf_exit_points[["uittredepunt_id", "afstand_buitenteenlijn"]]
    df_l_but = df_l_but.rename(columns={"afstand_buitenteenlijn": "mean"})
    append_to_gis_join_parameter_invoer_table(
        df_sjoin=df_l_but, parameter_name="L_but", geopackage_filepath=app_settings.geopackage_filepath)

    # Distance to binnenteenlijn
    gdf_binnenteenlijn: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="binnenteenlijn")
    gdf_exit_points['afstand_binnenteenlijn'] = gdf_exit_points.geometry.apply(
        lambda pnt: round(gdf_binnenteenlijn.distance(pnt).min(), 1))
    df_l_intrede = gdf_exit_points[["uittredepunt_id", "afstand_binnenteenlijn"]]
    df_l_intrede = df_l_intrede.rename(columns={"afstand_binnenteenlijn": "mean"})
    append_to_gis_join_parameter_invoer_table(
        df_sjoin=df_l_intrede, parameter_name="L_bit", geopackage_filepath=app_settings.geopackage_filepath)

    # Store back in geopackage
    gdf_exit_points.to_file(Path(app_settings.geopackage_filepath), layer="uittredepunten", driver="GPKG")
    print(BColors.OKBLUE,
          f"✅  Afstanden intrede, buitenteen en binnenteen zijn nu gekoppeld aan de uittredepunten.", BColors.ENDC)
    return True



from __future__ import annotations
import os
import sqlite3
from pandas import read_sql, read_excel
from typing import Optional, TYPE_CHECKING
from geoprob_pipe.cmd_app.parameter_input.initiate_input_excel_tables import DF_EMPTY_CORRELATIE_INVOER
from pandas import DataFrame
from geoprob_pipe.input_data.validation.dataframes.validation_objects import DataFrameQueryValidation
from geoprob_pipe.input_data.validation.dataframes.df_parameter_invoer import FAILURE_QUERIES
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def _load_df_correlatie_invoer_from_geopackage(geopackage_filepath: str) -> DataFrame:

    # Check if table exists in geopackage (older versions don't have this)
    conn = sqlite3.connect(geopackage_filepath)
    cursor = conn.cursor()
    cursor.execute(f"SELECT name FROM sqlite_master WHERE type='table' AND name='correlatie_invoer';")
    bool_table_exists = cursor.fetchone() is not None

    # Return empty if not exists
    if not bool_table_exists:
        return DF_EMPTY_CORRELATIE_INVOER

    # Return db table if exists
    df_correlatie_invoer = read_sql("SELECT * FROM correlatie_invoer;", conn)
    conn.close()
    return df_correlatie_invoer


def _validate_df_parameter_invoer(df: DataFrame, app_settings: ApplicationSettings):
    export_dir = os.path.join(
        os.path.dirname(app_settings.geopackage_filepath), "exports",
        str(app_settings.datetime_stamp), "parameter_input_process")

    obj = DataFrameQueryValidation(df=df, failure_queries=FAILURE_QUERIES)
    return obj.validate(export_dir=export_dir, label_humanized="Parameter invoer")


class InputParameterTables:

    def __init__(self, geopackage_filepath: str, path_to_excel: Optional[str] = None):

        # Placeholders
        self.df_scenario_invoer: Optional[DataFrame] = None
        self.df_parameter_invoer: Optional[DataFrame] = None
        self.df_fragility_values_invoer: Optional[DataFrame] = None
        self.df_correlatie_invoer: Optional[DataFrame] = None

        if path_to_excel is not None:
            self._load_data_from_excel(path_to_excel=path_to_excel, geopackage_filepath=geopackage_filepath)
        else:
            self._load_data_from_geopackage(geopackage_filepath=geopackage_filepath)

    def _load_data_from_geopackage(self, geopackage_filepath: str):
        conn = sqlite3.connect(geopackage_filepath)
        df_scenario_invoer = read_sql("SELECT * FROM scenario_invoer;", conn)
        self.df_scenario_invoer = df_scenario_invoer[["vak_id", "naam", "kans"]]
        self.df_parameter_invoer = read_sql("SELECT * FROM parameter_invoer;", conn)
        self.df_gis_join_parameter_invoer = read_sql("SELECT * FROM gis_join_parameter_invoer;", conn)
        self.df_fragility_values_invoer = read_sql("SELECT * FROM fragility_values_invoer;", conn)
        self.df_correlatie_invoer = _load_df_correlatie_invoer_from_geopackage(geopackage_filepath=geopackage_filepath)
        conn.close()

    def _load_data_from_excel(self, path_to_excel: str, geopackage_filepath: str):
        self.df_scenario_invoer = read_excel(path_to_excel, sheet_name="Scenario invoer", header=2)
        self.df_parameter_invoer = read_excel(path_to_excel, sheet_name="Parameter invoer", header=3)
        conn = sqlite3.connect(geopackage_filepath)
        self.df_gis_join_parameter_invoer = read_sql("SELECT * FROM gis_join_parameter_invoer;", conn)
        conn.close()
        self.df_fragility_values_invoer = read_excel(path_to_excel, sheet_name="Fragility values", header=3)
        self.df_correlatie_invoer = read_excel(path_to_excel, sheet_name="Correlatie invoer", header=3)

    def validate_and_report(self, app_settings: ApplicationSettings) -> bool:
        if not _validate_df_parameter_invoer(df=self.df_parameter_invoer, app_settings=app_settings): return False
        return True



from __future__ import annotations
from geoprob_pipe.calculations.systems.wbi.reliability_calculation import \
    WBICalculation
from geoprob_pipe.calculations.systems.base_objects.base_system_build import BaseSystemBuilder


class WBISystemBuilder(BaseSystemBuilder):

    def __init__(self,
                 geopackage_filepath: str,
                 to_run_vakken_ids: list[int]):
        super().__init__(geopackage_filepath=geopackage_filepath,
                         to_run_vakken_ids=to_run_vakken_ids)
        self.system_class = WBICalculation



from __future__ import annotations
from typing import TYPE_CHECKING
import os
from plotly.graph_objects import Figure as PlotlyFigure

from geoprob_pipe.visualizations.maps.betamap import BetaMap


if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


class Maps:

    def __init__(self, app_obj: GeoProbPipe):
        self.geoprob_pipe = app_obj

    @property
    def export_dir(self) -> str:
        path = os.path.join(self.geoprob_pipe.visualizations.export_dir, "maps")
        os.makedirs(path, exist_ok=True)
        return path

    def betamap(self, export: bool = False) -> PlotlyFigure:
        map_figure = BetaMap(self.geoprob_pipe, export=export)
        return map_figure.fig

    def export_maps(self):
        BetaMap(self.geoprob_pipe, export=True)



from __future__ import annotations
from geopandas import read_file
from shapely import LineString, MultiLineString, MultiPoint
from typing import TYPE_CHECKING
from geopandas import GeoDataFrame
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def load_dijktraject_linestring(app_settings: ApplicationSettings) -> LineString:
    gdf_dijktraject: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="dijktraject")
    gdf_dijktraject_geom = gdf_dijktraject.iloc[0].geometry
    if isinstance(gdf_dijktraject_geom, MultiLineString):
        assert gdf_dijktraject_geom.geoms.__len__() == 1
        ls_dijktraject: LineString = gdf_dijktraject_geom.geoms[0]
    elif isinstance(gdf_dijktraject_geom, LineString):
        ls_dijktraject = gdf_dijktraject_geom
    else:
        raise NotImplementedError(f"Type of '{type(gdf_dijktraject_geom)} is not yet supported. Please contact the "
                                  f"developer.'")
    return ls_dijktraject


def load_hydra_nl_as_multipoint(app_settings: ApplicationSettings) -> MultiPoint:
    gdf_hrd_locaties: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="hrd_locaties")
    return MultiPoint(gdf_hrd_locaties.geometry.values)



from __future__ import annotations
from geopandas import GeoDataFrame
from pandas import DataFrame
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


def _get_uittredepunten_gdf_beta_results(geoprob_pipe: GeoProbPipe, df_betas: DataFrame) -> GeoDataFrame:

    # Gather uittredepunten coord values
    gdf_uittredepunten = geoprob_pipe.input_data.uittredepunten.gdf
    gdf_coords = gdf_uittredepunten[["uittredepunt_id", "geometry"]]

    # Merge coord values to limit state dataframe
    df_merged = df_betas.merge(gdf_coords, left_on="uittredepunt_id", right_on="uittredepunt_id")

    return GeoDataFrame(df_merged, geometry="geometry", crs=gdf_uittredepunten.crs)


def get_gdf_beta_limit_states(geoprob_pipe: GeoProbPipe, export: bool = False) -> GeoDataFrame:
    gdf = _get_uittredepunten_gdf_beta_results(geoprob_pipe, geoprob_pipe.results.df_beta_limit_states)
    if export:
        gdf.to_file(geoprob_pipe.input_data.app_settings.geopackage_filepath,
                    layer="beta_limit_states", driver="GPKG", mode="w")
    return gdf


def get_gdf_beta_scenarios_final(geoprob_pipe: GeoProbPipe, export: bool = False) -> GeoDataFrame:
    gdf = _get_uittredepunten_gdf_beta_results(geoprob_pipe, geoprob_pipe.results.df_beta_scenarios_final)
    if export:
        gdf.to_file(
            geoprob_pipe.input_data.app_settings.geopackage_filepath,
            layer="beta_scenarios_final", driver="GPKG", mode="w")
    return gdf


def get_gdf_beta_scenarios_cp(geoprob_pipe: GeoProbPipe, export: bool = False) -> GeoDataFrame:
    gdf = _get_uittredepunten_gdf_beta_results(geoprob_pipe, geoprob_pipe.results.df_beta_scenarios_cp)
    if export:
        gdf.to_file(
            geoprob_pipe.input_data.app_settings.geopackage_filepath,
            layer="beta_scenarios_cp", driver="GPKG", mode="w")
    return gdf


def get_gdf_beta_scenarios_rp(geoprob_pipe: GeoProbPipe, export: bool = False) -> GeoDataFrame:
    gdf = _get_uittredepunten_gdf_beta_results(geoprob_pipe, geoprob_pipe.results.df_beta_scenarios_rp)
    if export:
        gdf.to_file(
            geoprob_pipe.input_data.app_settings.geopackage_filepath,
            layer="beta_scenarios_rp", driver="GPKG", mode="w")
    return gdf


def get_gdf_beta_uittredepunten(geoprob_pipe: GeoProbPipe, export: bool = False) -> GeoDataFrame:
    gdf = _get_uittredepunten_gdf_beta_results(geoprob_pipe, geoprob_pipe.results.df_beta_uittredepunten)
    if export:
        gdf.to_file(
            geoprob_pipe.input_data.app_settings.geopackage_filepath,
            layer="beta_uittredepunten", driver="GPKG", mode="w")
    return gdf



from geoprob_pipe.cmd_app.spatial_layers.dijktraject import added_dijktraject
from geoprob_pipe.cmd_app.spatial_layers.vakindeling import added_vakindeling
from geoprob_pipe.cmd_app.spatial_layers.uittredepunten.uittredepunten import added_uittredepunten
from geoprob_pipe.cmd_app.spatial_layers.hrd import added_hrd
from geoprob_pipe.cmd_app.spatial_layers.polderpeil import added_polderpeil
from geoprob_pipe.cmd_app.spatial_layers.binnenteenlijn import added_binnenteenlijn
from geoprob_pipe.cmd_app.spatial_layers.buitenteenlijn import added_buitenteenlijn
from geoprob_pipe.cmd_app.spatial_layers.intredelijn import added_intredelijn
from geoprob_pipe.cmd_app.spatial_layers.ahn import added_ahn



from __future__ import annotations
import os
from typing import Union, Literal
from pandas import DataFrame, Series
from copy import deepcopy
# from geoprob_pipe.utils.other import repository_root_path
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe



class VisualizeInfo:

    def __init__(self, main_tag: str, beta: Union[float, int, str] = "n.b.", visible: bool = False):
        self.main_tag: str = main_tag
        self.beta: Union[float, int, str] = beta
        self.visible: bool = visible

    @property
    def svg_color(self) -> str:
        """ Color value formatted for the SVG-file. """
        if self.beta > 5.00:
            return "rgb(154,205,50)"  # Green
        return "rgb(206,32,41)"  # Red

    @property
    def svg_beta(self) -> str:
        """ Beta value formatted for the SVG-file. """
        if isinstance(self.beta, float) or isinstance(self.beta, int):
            if self.beta >= 10.0:
                return f"{round(self.beta, 1):.1f}"
            return f"{round(self.beta, 2):.2f}"
        return str(self.beta)

    @property
    def svg_visibility(self):
        """ Visibility value formatted for the SVG-file. """
        if self.visible:
            return "visible"
        return "hidden"

    def svg_tag(self, option: Literal["beta", "color", "visibility"]) -> str:
        return f"{self.main_tag}.{option}"

    def svg_tag_incl_brackets(self, option: Literal["beta", "color", "visibility"]) -> str:
        return f"{{{{ {self.svg_tag(option=option)} }}}}"


DICT_VISUALIZE_INFO = {
    "uplift": VisualizeInfo(main_tag="uplift", beta=-1, visible=True),
    "heave": VisualizeInfo(main_tag="heave", beta=-1, visible=True),
    "piping": VisualizeInfo(main_tag="piping", beta=-1, visible=True),
    "scenario.1": VisualizeInfo(main_tag="scenario.1", beta=-1),
    "scenario.2": VisualizeInfo(main_tag="scenario.2", beta=-1),
    "scenario.3": VisualizeInfo(main_tag="scenario.3", beta=-1),
    "scenario.4": VisualizeInfo(main_tag="scenario.4", beta=-1),
    "scenario.5": VisualizeInfo(main_tag="scenario.5", beta=-1),
    "scenario.6": VisualizeInfo(main_tag="scenario.6", beta=-1),
    "scenario.7": VisualizeInfo(main_tag="scenario.7", beta=-1),
    "scenario.8": VisualizeInfo(main_tag="scenario.8", beta=-1),
    "scenario.9": VisualizeInfo(main_tag="scenario.9", beta=-1),
    "scenario.10": VisualizeInfo(main_tag="scenario.10", beta=-1),
    "scenario.11": VisualizeInfo(main_tag="scenario.11", beta=-1),
    "scenario.12": VisualizeInfo(main_tag="scenario.12", beta=-1),
    "scenario.13": VisualizeInfo(main_tag="scenario.13", beta=-1),
    "scenario.14": VisualizeInfo(main_tag="scenario.14", beta=-1),
    "scenario.15": VisualizeInfo(main_tag="scenario.15", beta=-1),
    "scenario.16": VisualizeInfo(main_tag="scenario.16", beta=-1),
    "scenario.17": VisualizeInfo(main_tag="scenario.17", beta=-1),
    "scenario.18": VisualizeInfo(main_tag="scenario.18", beta=-1),
    "scenario.19": VisualizeInfo(main_tag="scenario.19", beta=-1),
    "scenario.20": VisualizeInfo(main_tag="scenario.20", beta=-1),
    "scenario.99": VisualizeInfo(main_tag="scenario.99", beta=-1),
    "uittredepunt.1": VisualizeInfo(main_tag="uittredepunt.1", beta=-1),
    "uittredepunt.2": VisualizeInfo(main_tag="uittredepunt.2", beta=-1),
    "uittredepunt.3": VisualizeInfo(main_tag="uittredepunt.3", beta=-1),
    "uittredepunt.4": VisualizeInfo(main_tag="uittredepunt.4", beta=-1),
    "uittredepunt.5": VisualizeInfo(main_tag="uittredepunt.5", beta=-1),
    "uittredepunt.6": VisualizeInfo(main_tag="uittredepunt.6", beta=-1),
    "uittredepunt.7": VisualizeInfo(main_tag="uittredepunt.7", beta=-1),
    "uittredepunt.8": VisualizeInfo(main_tag="uittredepunt.8", beta=-1),
    "uittredepunt.9": VisualizeInfo(main_tag="uittredepunt.9", beta=-1),
    "uittredepunt.10": VisualizeInfo(main_tag="uittredepunt.10", beta=-1),
    "uittredepunt.11": VisualizeInfo(main_tag="uittredepunt.11", beta=-1),
    "uittredepunt.12": VisualizeInfo(main_tag="uittredepunt.12", beta=-1),
    "uittredepunt.13": VisualizeInfo(main_tag="uittredepunt.13", beta=-1),
    "uittredepunt.14": VisualizeInfo(main_tag="uittredepunt.14", beta=-1),
    "uittredepunt.15": VisualizeInfo(main_tag="uittredepunt.15", beta=-1),
    "uittredepunt.16": VisualizeInfo(main_tag="uittredepunt.16", beta=-1),
    "uittredepunt.17": VisualizeInfo(main_tag="uittredepunt.17", beta=-1),
    "uittredepunt.18": VisualizeInfo(main_tag="uittredepunt.18", beta=-1),
    "uittredepunt.19": VisualizeInfo(main_tag="uittredepunt.19", beta=-1),
    "uittredepunt.20": VisualizeInfo(main_tag="uittredepunt.20", beta=-1),
    "uittredepunt.21": VisualizeInfo(main_tag="uittredepunt.21", beta=-1),
    "uittredepunt.99": VisualizeInfo(main_tag="uittredepunt.99", beta=-1),
    "vak.1": VisualizeInfo(main_tag="vak.1", beta=-1),
    "vak.2": VisualizeInfo(main_tag="vak.2", beta=-1),
    "vak.3": VisualizeInfo(main_tag="vak.3", beta=-1),
    "vak.4": VisualizeInfo(main_tag="vak.4", beta=-1),
    "vak.5": VisualizeInfo(main_tag="vak.5", beta=-1),
    "vak.6": VisualizeInfo(main_tag="vak.6", beta=-1),
    "vak.7": VisualizeInfo(main_tag="vak.7", beta=-1),
    "vak.8": VisualizeInfo(main_tag="vak.8", beta=-1),
    "vak.9": VisualizeInfo(main_tag="vak.9", beta=-1),
    "vak.10": VisualizeInfo(main_tag="vak.10", beta=-1),
    "vak.11": VisualizeInfo(main_tag="vak.11", beta=-1),
    "vak.12": VisualizeInfo(main_tag="vak.12", beta=-1),
    "vak.13": VisualizeInfo(main_tag="vak.13", beta=-1),
    "vak.14": VisualizeInfo(main_tag="vak.14", beta=-1),
    "vak.15": VisualizeInfo(main_tag="vak.15", beta=-1),
    "vak.16": VisualizeInfo(main_tag="vak.16", beta=-1),
    "vak.17": VisualizeInfo(main_tag="vak.17", beta=-1),
    "vak.18": VisualizeInfo(main_tag="vak.18", beta=-1),
    "vak.19": VisualizeInfo(main_tag="vak.19", beta=-1),
    "vak.20": VisualizeInfo(main_tag="vak.20", beta=-1),
    "vak.21": VisualizeInfo(main_tag="vak.21", beta=-1),
    "vak.99": VisualizeInfo(main_tag="vak.99", beta=-1),
    "traject": VisualizeInfo(main_tag="traject", beta=-1),
}


def populate_visualize_dict(uittredepunt_id: int, ondergrondscenario_id: int, app_obj: GeoProbPipe):

    visualize_dict = deepcopy(DICT_VISUALIZE_INFO)

    # Populate uplift, heave and piping for focus scenario
    df_filter_limit_states = app_obj.results.df_limit_states.copy(deep=True)
    df_filter_limit_states: DataFrame = df_filter_limit_states[
        (df_filter_limit_states["uittredepunt_id"] == uittredepunt_id) &
        (df_filter_limit_states["ondergrondscenario_id"] == ondergrondscenario_id)
    ]
    assert df_filter_limit_states.__len__() == 3
    for row in df_filter_limit_states.itertuples(index=False):
        row: Series
        visualize_dict[row.model].beta = row.beta

    # Populate scenarios results: other scenarios
    df_filter_combined = app_obj.results.df_combined.copy(deep=True)
    df_filter_combined: DataFrame = df_filter_combined[
        (df_filter_combined["uittredepunt_id"] == uittredepunt_id) &
        (df_filter_combined["ondergrondscenario_id"] != ondergrondscenario_id)
    ]
    for index, row in enumerate(df_filter_combined.itertuples(index=False)):
        row: Series
        visualize_dict[f"scenario.{index+2}"].beta = row.beta
        visualize_dict[f"scenario.{index+2}"].visible = True

    # Populate scenarios results: focus scenarios
    df_filter_combined = app_obj.results.df_combined.copy(deep=True)
    df_filter_combined: DataFrame = df_filter_combined[
        (df_filter_combined["uittredepunt_id"] == uittredepunt_id) &
        (df_filter_combined["ondergrondscenario_id"] == ondergrondscenario_id)
    ]
    assert df_filter_combined.__len__() == 1
    for row in df_filter_combined.itertuples(index=False):
        row: Series
        visualize_dict["scenario.1"].beta = row.beta
        visualize_dict["scenario.1"].visible = True

    # Populate vak

    return visualize_dict


def generate_overview_flow_chart_with_betas(
        uittredepunt_id: int, ondergrondscenario_id: int, app_obj: GeoProbPipe, export_dir: str
):
    """ Generates a flow chart that provides an overview what the beta values are per step in the calculation process.
    It displays it from the given scenario and uittredepunt, until vak- and traject-level. """

    dict_to_use = populate_visualize_dict(
        uittredepunt_id=uittredepunt_id, ondergrondscenario_id=ondergrondscenario_id, app_obj=app_obj)

    # Read template
    svg_text = None
    # repo_root = repository_root_path()
    path_to_svg = os.path.join(
        repo_root, "geoprob_pipe", "graphs", "overview", "Hierarchie_berekeningen_incl_result_tags_v2.svg")
    # TODO: Use 'import importlib.resources' for this
    with open(path_to_svg, "r", encoding="utf-8") as f:
        svg_text = f.read()
    if svg_text is None:
        raise ValueError

    # Replace tags
    for main_tag, info in dict_to_use.items():
        svg_text = svg_text.replace(info.svg_tag_incl_brackets(option="beta"), info.svg_beta)
        svg_text = svg_text.replace(info.svg_tag_incl_brackets(option="color"), info.svg_color)
        svg_text = svg_text.replace(info.svg_tag_incl_brackets(option="visibility"), info.svg_visibility)

    # Save new svg
    with open(os.path.join(export_dir, "results_overview_flow_chart.svg"), "w", encoding="utf-8") as f:
        f.write(svg_text)



from __future__ import annotations
from pandas import DataFrame, merge
import numpy as np
import os
from datetime import datetime
import plotly.graph_objects as go
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


def _add_beta_per_uittredepunt_points(
        self, df_for_graph: DataFrame, mask, name: str, color: str, value: bool):
    """ Visualization of beta of uittredepunten.

    :param self: GraphBetaValuesSingleInteractive-object
    :param df_for_graph: Data to plot
    :param mask: Boolean Series for which uittredepunten to plot in graph. Either converged of unconverged points.
    :param name: Marker name
    :param color: Marker color
    :param value: Requested option: converged or not converged
    """

    self.fig.add_trace(
        go.Scatter(
            x=df_for_graph.loc[mask, "metrering"], y=df_for_graph.loc[mask, "beta"],
            mode='markers', marker=dict(symbol='circle', size=7, color=color), name=name,
            customdata=df_for_graph.loc[mask, ["uittredepunt_id", "beta", "metrering"]],
            hovertemplate=("ID: %{customdata[0]}<br>" +
                           "Beta: %{customdata[1]:.3f}<br>" +
                           "Metrering: %{customdata[2]}"),
            legendgroup="Beta uittredepunten", showlegend=value))


def _add_beta_per_uittredepunt_indication_above_plotting_range(
        self, df_for_graph: DataFrame, mask, name: str, color: str, value: bool):
    """ Indication to user that there are Beta results plotted outside the plotting range. In this case above range. """

    self.fig.add_trace(
        go.Scatter(
            x=df_for_graph.loc[mask, 'metrering'],
            y=[self.beta_max - 0.1] * mask.sum(),
            mode='markers',
            marker=dict(symbol='triangle-up', size=7, color=color),
            name=name + " above plotted range",
            customdata=df_for_graph.loc[mask, ["uittredepunt_id", "beta", "metrering"]],
            hovertemplate=("ID: %{customdata[0]}<br>" +
                           "Beta: %{customdata[1]:.3f}<br>" +
                           "Metrering: %{customdata[2]}"),
            legendgroup="Beta uittredepunten",
            showlegend=value
        )
    )


def _add_beta_per_uittredepunt_indication_below_plotting_range(
        self, df_for_graph: DataFrame, mask, name: str, color: str, value: bool):
    """ Indication to user that there are Beta results plotted outside the plotting range. In this case below range. """

    self.fig.add_trace(
        go.Scatter(
            x=df_for_graph.loc[mask, 'metrering'],
            y=[self.beta_min + 0.1] * mask.sum(),
            mode='markers',
            marker=dict(symbol='triangle-down', size=7, color=color),
            name=name + " below plotted range",
            customdata=df_for_graph.loc[mask, ["uittredepunt_id", "beta", "metrering"]],
            hovertemplate=("ID: %{customdata[0]}<br>" +
                           "Beta: %{customdata[1]:.3f}<br>" +
                           "Metrering: %{customdata[2]}"),
            legendgroup="Beta uittredepunten",
            showlegend=value
        )
    )


class GraphBetaValuesSingleInteractive:

    def __init__(self, geoprob_pipe: GeoProbPipe, export: bool = True):

        # Logic
        self.geoprob_pipe = geoprob_pipe
        self.fig = go.Figure()

        self.gdf_uittredepunten = self.geoprob_pipe.input_data.uittredepunten.gdf
        self.gdf_vakken = self.geoprob_pipe.input_data.vakken.gdf
        self.m_start = self.gdf_vakken['m_start'].min()-10
        self.m_end = self.gdf_vakken['m_end'].max()+10

        self.beta_min = 2
        self.beta_max = 20

        self._add_backgrond()
        self._add_beta_per_vak()
        self._add_beta_per_scenario()
        self._add_beta_per_uittredepunt()
        self._update_layout()
        self._optionally_export(export=export)

    def _add_backgrond(self):
        # Oude categorie grenzen
        # cg = (self.geoprob_pipe.input_data.traject_normering
        #       .beta_categorie_grenzen)
        # colors = ["rgba(0,128,0,0.4)", "rgba(144,238,144,0.4)",
        #           "rgba(255,255,0,0.4)", "rgba(255,165,0,0.4)",
        #           "rgba(255,0,0,0.4)", "rgba(128,0,128,0.4)"]

        # labels = ["β<sub>eis;sig;dsn / 30</sub>", "β<sub>eis;sig;dsn</sub>",
        #           "β<sub>eis;ond;dsn</sub>", "β<sub>eis;ond</sub>",
        #           "β<sub>eis;ond * 30</sub>", ""]

        cg = self.geoprob_pipe.input_data.traject_normering.riskeer_categorie_grenzen
        colors = ["rgba(30,141,41,0.6)", "rgba(146,206,90,0.6)",
                  "rgba(198,226,176,0.6)", "rgba(255,255,0,0.6)",
                  "rgba(254,165,3,0.6)", "rgba(255,0,0,0.6)",
                  "rgba(177,33,38,0.6)"]
        labels = ["+III", "+II", "+I", "0", "-I", "-II", "-III"]

        x_line = np.linspace(self.m_start, self.m_end)
        self.annotation_vak = []
        self.annotation_vak.append(dict(x=0.5,
                                        y=np.log10(2.1),
                                        text="Vak ID:",
                                        showarrow=False,
                                        xanchor="left",
                                        yanchor="bottom",
                                        font=dict(color="black")
                                        ))
        self.vak_lines = []
        for _, vak in self.gdf_vakken.iterrows():
            self.vak_lines.append(dict(
                x0=vak["m_start"], x1=vak["m_start"],
                y0=0, y1=1,
                xref="x", yref="paper",
                line=dict(color="black", width=1)
                ))
            self.vak_lines.append(dict(
                x0=vak["m_end"], x1=vak["m_end"],
                y0=0, y1=1,
                xref="x", yref="paper",
                line=dict(color="black", width=1)
                ))
            self.annotation_vak.append(dict(
                x=(vak["m_start"] + vak["m_end"]) / 2, y=np.log10(2),
                text=vak["id"],
                showarrow=False,
                xanchor="center",
                yanchor="bottom",
                font=dict(color="black")
                ))
        self.annotation_label = []
        for i, grens in enumerate(cg):

            if cg[grens][0] <= 0:
                cg[grens][0] = np.log10(2)

            # Onderste lijn (zichtbaar)
            self.fig.add_trace(
                go.Scatter(
                    x=x_line,
                    y=[cg[grens][0]] * len(x_line),
                    name=grens,
                    mode="lines",
                    line=dict(color="black", width=0.5),
                    hoverinfo="skip",
                    showlegend=False,
                )
            )

            # Bovenste lijn (onzichtbaar, zorgt voor fill)
            self.fig.add_trace(go.Scatter(
                x=x_line,
                y=[cg[grens][1]] * len(x_line),
                name=grens,
                mode="lines",
                line=dict(width=0),        # geen bovenrand zichtbaar
                fill="tonexty",
                fillcolor=colors[i % len(colors)],  # kleur uit lijst
                hoverinfo="skip",
                showlegend=False,
            ))

            # Labels bij de ondergrens
            self.annotation_label.append(dict(
                x=x_line.max(),
                y=(np.log10(cg[grens][0]) + np.log10(cg[grens][1])) / 2,
                text=labels[i % len(labels)],
                showarrow=False,
                xanchor="left",
                yanchor="middle",
                font=dict(color="black", size=10),
                align="right"
            ))

    def _add_beta_per_scenario(self):
        df_beta_scenarios_final = self.geoprob_pipe.results.df_beta_scenarios_final
        df_for_graph = merge(
            left=df_beta_scenarios_final[[
                "uittredepunt_id", "beta", "converged", "method_used", "flow_chart_number", "advise"]],
            right=self.gdf_uittredepunten[["uittredepunt_id", "metrering"]],
            on="uittredepunt_id", how="left")
        df_for_graph["converged_alt"] = df_for_graph["flow_chart_number"] != 5

        for value, color, name in [(True, "black", 'Beta scenarios'),
                                   (False, "blue", "Verder rekenen geadviseerd (scenarios)")]:
            mask = df_for_graph["converged_alt"] == value
            mask_low = df_for_graph["beta"] < self.beta_min
            mask_high = df_for_graph["beta"] > self.beta_max

            # In range
            self.fig.add_trace(
                go.Scatter(
                    x=df_for_graph.loc[mask, 'metrering'], y=df_for_graph.loc[mask, "beta"], mode='markers',
                    marker=dict(symbol='diamond', size=7, color=color, line=dict(color="white", width=1)),
                    legendgroup="Beta scenarios", showlegend=value, name=name,
                    customdata=df_for_graph.loc[mask, [
                        "uittredepunt_id", "beta", "metrering", "method_used", "converged", "advise"]],
                    hovertemplate=(
                        "ID: %{customdata[0]}<br>"
                        "Beta: %{customdata[1]:.3f}<br>"
                        "Metrering: %{customdata[2]}<br>"
                        "Methode: %{customdata[3]}<br>"
                        "Converged: %{customdata[4]}<br>"
                        "Advies: %{customdata[5]}")))

            # Above range
            mask_combi = mask & mask_high
            self.fig.add_trace(
                go.Scatter(
                    x=df_for_graph.loc[mask_combi, 'metrering'], y=[self.beta_max-0.1] * mask_combi.sum(),
                    mode='markers',
                    marker=dict(symbol='triangle-up', size=7, color=color, line=dict(color="white", width=1)),
                    legendgroup="Beta scenarios", showlegend=value, name=name + " above plotted range",
                    customdata=df_for_graph.loc[mask_combi, ["uittredepunt_id", "beta", "metrering"]],
                    hovertemplate="ID: %{customdata[0]}<br>"
                                  "Beta: %{customdata[1]:.3f}<br>"
                                  "Metrering: %{customdata[2]}"))

            # Below range
            mask_combi = mask & mask_low
            self.fig.add_trace(
                go.Scatter(
                    x=df_for_graph.loc[mask_combi, 'metrering'], y=[self.beta_min+0.1] * mask_combi.sum(),
                    mode='markers', name=name + " below plotted range",
                    marker=dict(symbol='triangle-down', size=7, color=color, line=dict(color="white", width=1)),
                    legendgroup="Beta scenarios", showlegend=value,
                    customdata=df_for_graph.loc[mask_combi, ["uittredepunt_id", "beta", "metrering"]],
                    hovertemplate=("ID: %{customdata[0]}<br>" +
                                   "Beta: %{customdata[1]:.3f}<br>" +
                                   "Metrering: %{customdata[2]}")))

    def _add_beta_per_uittredepunt(self):

        # Gather results to plot
        df_results_uittredepunten = self.geoprob_pipe.results.df_beta_uittredepunten
        df_for_graph: DataFrame = merge(
            left=df_results_uittredepunten[["uittredepunt_id", "beta", "converged", "flow_chart_number", "advise"]],
            right=self.gdf_uittredepunten[["uittredepunt_id", "metrering"]],
            on="uittredepunt_id",
            how="left"
        )
        df_for_graph["converged_alt"] = df_for_graph["flow_chart_number"] != 11

        # Plot converged and 'Verder rekenen geadviseerd' values
        for value, color, name in [
            (True, "black", "Beta uittredepunt"),
            (False, "blue", "Verder rekenen geadviseerd (uittredepunten)")
        ]:

            mask = df_for_graph["converged_alt"] == value
            _add_beta_per_uittredepunt_points(
                self=self, df_for_graph=df_for_graph, mask=mask, name=name, color=color, value=value)

            mask_high = df_for_graph["beta"] > self.beta_max
            _add_beta_per_uittredepunt_indication_above_plotting_range(
                self=self, df_for_graph=df_for_graph, mask=mask & mask_high, name=name, color=color, value=value)

            mask_low = df_for_graph["beta"] < self.beta_min
            _add_beta_per_uittredepunt_indication_below_plotting_range(
                self=self, df_for_graph=df_for_graph, mask=mask & mask_low, name=name, color=color, value=value)

    def _add_beta_per_vak(self):

        # Gather data
        df_results_vakken = self.geoprob_pipe.results.df_beta_vakken
        df_results_vakken = df_results_vakken.rename(columns={"vak_id": "id"})
        df_for_graph = merge(
            left=df_results_vakken[["id", "beta", "advise"]],
            right=self.gdf_vakken[["id", "m_start", "m_end"]],
            on="id", how="left")
        first = True

        # Iterate and plot
        for value, color, name in [(True, "black", "Beta vakken"),
                                   (False, "blue", "Verder rekenen geadviseerd (vakken)")]:

            # Plotting vak lines
            mask = (df_for_graph["advise"] == "-") == value
            mask_low = df_for_graph["beta"] < self.beta_min
            mask_high = df_for_graph["beta"] > self.beta_max
            for _, row in df_for_graph.loc[mask].iterrows():
                self.fig.add_trace(go.Scatter(
                    x=[row["m_start"], row["m_end"]], y=[row["beta"], row["beta"]],
                    mode="lines", line=dict(color=color, width=2.5), name=name, legendgroup="Beta vakken",
                    showlegend=first & value, customdata=[[row["id"], row["beta"]]] * 2,
                    hovertemplate="ID: %{customdata[0]}<br>" +
                                  "Beta: %{customdata[1]:.3f}"))
                first = False

            # Above range
            mask_combi = mask & mask_high
            first = True
            for _, row in df_for_graph.loc[mask_combi].iterrows():
                self.fig.add_trace(go.Scatter(
                    x=[(row["m_start"] + row["m_end"]) / 2],
                    y=[self.beta_max-0.1],
                    mode="markers",
                    marker=dict(color=color, symbol="triangle-up", size=9),
                    name=name + " above plotted range",
                    customdata=[[row["id"], row["beta"]]],
                    hovertemplate=("ID: %{customdata[0]}<br>" +
                                   "Beta: %{customdata[1]:.3f}<br>"),
                    legendgroup="Beta vakken",
                    showlegend=first & value
                ))
                first = False

            # Below range
            mask_combi = mask & mask_low
            first = True
            for _, row in df_for_graph.loc[mask_combi].iterrows():
                self.fig.add_trace(go.Scatter(
                    x=[(row["m_start"] + row["m_end"]) / 2], y=[self.beta_min+0.1],
                    mode="markers", marker=dict(color=color, symbol="triangle-down", size=9),
                    name=name + " below plotted range",
                    customdata=[[row["id"], row["beta"]]],
                    hovertemplate="ID: %{customdata[0]}<br>" +
                                  "Beta: %{customdata[1]:.3f}<br>",
                    legendgroup="Beta vakken", showlegend=first & value
                ))
                first = False

    def _update_layout(self):
        self.fig.add_trace(go.Scatter(
            x=[0], y=[0], name="Verder rekenen geadviseerd", showlegend=True,
            mode="markers", marker=dict(color="blue", symbol="square")))
        annotation = self.annotation_vak + self.annotation_label
        self.fig.update_layout(
            title="Betrouwbaarheidsindex STPH",
            xaxis=dict(title="Metrering",
                       type='linear',
                       range=[0, self.m_end],
                       showgrid=True,
                       gridwidth=0.5,
                       gridcolor="gray"
                       ),
            yaxis=dict(title="Betrouwbaarheidsindex β [-]",
                       type='log',
                       range=[np.log10(2), np.log10(20)],
                       showgrid=True,
                       gridwidth=0.5,
                       gridcolor="gray",
                       tickmode="array",
                       tickvals=[2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 20],
                       ticktext=[2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 20]
                       ),
            legend=dict(
                yanchor="top",
                y=0.99,
                xanchor="left",
                x=0.01
                ),
            annotations=annotation,
            shapes=self.vak_lines
            )
        # Toggles for annotations
        self.fig.update_layout(
            updatemenus=[
                dict(
                    type="buttons",
                    direction="right",
                    buttons=[
                        dict(
                            label="Show annotations",
                            method="relayout",
                            args=["annotations",
                                  self.annotation_vak + self.annotation_label]
                        ),
                        dict(
                            label="Hide annotations",
                            method="relayout",
                            args=["annotations", self.annotation_label]
                        ),
                        dict(
                            label="Show vlines",
                            method="relayout",
                            args=["shapes", self.vak_lines]
                        ),
                        dict(
                            label="Hide vlines",
                            method="relayout",
                            args=["shapes", []]
                        )
                    ],
                    x=0.5,
                    y=1.15,
                    xanchor="center"
                )
            ]
        )

    def _optionally_export(self, export: bool = False, add_timestamp: bool = False):
        if not export:
            return
        os.makedirs(self.geoprob_pipe.visualizations.graphs.export_dir, exist_ok=True)
        timestamp_str = ""
        if add_timestamp:
            timestamp = datetime.now().strftime("%Y-%m-%d_%H%M")
            timestamp_str = f"{timestamp}_"
        self.fig.write_html(os.path.join(
            self.geoprob_pipe.visualizations.graphs.export_dir, f"{timestamp_str}betrouwbaarheidsindex.html"
            ), include_plotlyjs='cdn')



from __future__ import annotations
from typing import TYPE_CHECKING
from shapely import MultiLineString, LineString
from geoprob_pipe.utils.validation_messages import BColors
from geopandas import GeoDataFrame, read_file
from pathlib import Path
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def coupled_uittredepunten_to_refline(app_settings: ApplicationSettings) -> bool:
    """ Controleert of de metrering al gekoppeld is aan de uittredepunten. Indien dit niet het geval is, doet deze
    functie dat automatisch. """

    # Read uittredepunten
    gdf_exit_points: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="uittredepunten")

    # Check if already added
    if "metrering" in gdf_exit_points.columns and "afstand_reflijn" in gdf_exit_points.columns:
        print(BColors.OKBLUE, f"✔  Afstand en metrering tot reflijn al gekoppeld aan uittredepunten.", BColors.ENDC)
        return True  # Assuming already added

    gdf_dijktraject: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="dijktraject")
    gdf_dijktraject_geom = gdf_dijktraject.iloc[0].geometry
    if isinstance(gdf_dijktraject_geom, MultiLineString):
        assert gdf_dijktraject_geom.geoms.__len__() == 1
        ls_dijktraject: LineString = gdf_dijktraject_geom.geoms[0]
    elif isinstance(gdf_dijktraject_geom, LineString):
        ls_dijktraject: LineString = gdf_dijktraject_geom
    else:
        raise NotImplementedError(f"Type of {type(gdf_dijktraject_geom)} is not yet implemented.")

    # Spatial analyses
    gdf_exit_points['metrering'] = gdf_exit_points.geometry.apply(
        lambda pnt: round(ls_dijktraject.project(pnt), 1))
    gdf_exit_points['afstand_reflijn'] = gdf_exit_points.geometry.apply(
        lambda pnt: round(ls_dijktraject.distance(pnt), 1))

    # Add uittredepunt id
    gdf_exit_points = gdf_exit_points.sort_values(by=["metrering"])
    gdf_exit_points['uittredepunt_id'] = range(1, len(gdf_exit_points) + 1)

    # Store back in geopackage
    gdf_exit_points.to_file(Path(app_settings.geopackage_filepath), layer="uittredepunten", driver="GPKG")
    print(BColors.OKBLUE,
          f"✅  Afstand en metrering tot reflijn zijn nu gekoppeld aan de uittredepunten.", BColors.ENDC)
    return True



from __future__ import annotations
from geoprob_pipe.calculations.systems.base_objects.system_calculation import SystemCalculation
from typing import List, Tuple
from pandas import DataFrame, Series
import sqlite3
from geoprob_pipe.cmd_app.parameter_input.expand_input_tables import run_expand_input_tables


def _gather_variable_correlations(geopackage_filepath: str
                                  ) -> List[Tuple[str, str, float]]:
    """ Input originally from the input Excel-file. It defines per combination of two parameters the correlation; a
    value between 0.0 (no correlation) and 1.0 (fully correlated). By default, no correlation is specified for
    parameter combinations, i.e. the probabilistic library automatically assigns 0.0 as correlation for the
    combination. The correlation applies for the entire trajectory.

    TODO: This setup applies for the entire trajectory. In future versions of
    the code the user should be able to assign the correlation on vak, scenario and uittredepunten level.

    :return:
    """

    try:
        conn = sqlite3.connect(geopackage_filepath)
        cursor = conn.cursor()
        cursor.execute("""
SELECT parameter_a, parameter_b, correlation FROM correlatie_invoer
WHERE correlation <> 0.0;
""")
        rows = cursor.fetchall()  # This will be a list of tuples
        conn.close()
        return rows
    except sqlite3.OperationalError:
        return []


def _gather_calculation_input(
        df_expanded: DataFrame, uittredepunt_id: int,
        ondergrondscenario_naam: str) -> List:

    # Collect input for specific calculation
    df_filter = df_expanded[
        (df_expanded["uittredepunt_id"] == uittredepunt_id) &
        (df_expanded["ondergrondscenario_naam"] == ondergrondscenario_naam)]
    df = df_filter.copy(deep=True)

    # Parse parameter input to list of dictionaries
    df['parameter_input'] = df.apply(
        lambda row2: {**row2['parameter_input'], 'name': row2['parameter_name']}, axis=1)
    return df['parameter_input'].values.tolist()


def _generate_single_calculation(
        row_calculation_metadata: Series, vak_id: int, df_expanded: DataFrame, system_class,
        variable_correlations: List[Tuple[str, str, float]]
) -> SystemCalculation:

    # General information
    uittredepunt_id = row_calculation_metadata["uittredepunt_id"]
    ondergrondscenario_naam = row_calculation_metadata["ondergrondscenario_naam"]

    # Construct calculation
    calculation_input = _gather_calculation_input(
        df_expanded=df_expanded, uittredepunt_id=uittredepunt_id, ondergrondscenario_naam=ondergrondscenario_naam)
    calc: SystemCalculation = system_class(distributions=calculation_input, correlations=variable_correlations)
    calc.metadata["uittredepunt_id"] = uittredepunt_id
    calc.metadata["ondergrondscenario_naam"] = ondergrondscenario_naam
    calc.metadata["vak_id"] = vak_id
    metadata_summary = {
        "uittredepunt_id": uittredepunt_id,
        "ondergrondscenario_naam": ondergrondscenario_naam,
        "vak_id": vak_id}
    calc.validation_messages.about = f"Calculation {metadata_summary}"

    return calc


class BaseSystemBuilder:

    def __init__(self, geopackage_filepath: str, to_run_vakken_ids: list[int]):
        self.system_class = SystemCalculation
        self.geopackage_filepath = geopackage_filepath

        # Gather input
        df_expanded = run_expand_input_tables(geopackage_filepath=self.geopackage_filepath)

        # Filter vakken (if only selection needs to run)
        if to_run_vakken_ids is not None:
            df_expanded = df_expanded[df_expanded['vak_id'].isin(to_run_vakken_ids)]
        self.df_expanded = df_expanded

    def setup_iteration_df(self) -> DataFrame:
        # Iteration dataframe
        df_unique_combos: DataFrame = self.df_expanded[[
            "uittredepunt_id", "ondergrondscenario_naam", "vak_id"]].drop_duplicates()
        return df_unique_combos

    def build_instance(self, row_unique) -> SystemCalculation:

        # Gather variable correlations
        variable_correlations: List[Tuple[str, str, float]] = _gather_variable_correlations(self.geopackage_filepath)
        # TODO: Should be made uittredepunt/vak specific in future versions
        # of the code. For now only for the entire trajectory.

        calc = _generate_single_calculation(
            row_calculation_metadata=row_unique, vak_id=row_unique["vak_id"], df_expanded=self.df_expanded,
            system_class=self.system_class, variable_correlations=variable_correlations)

        return calc



from __future__ import annotations
from typing import TYPE_CHECKING
import plotly.graph_objects as go
import os
import geopandas as gpd
from shapely.geometry import LineString, MultiLineString, GeometryCollection

if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


def _add_line(geoprob_pipe: GeoProbPipe, fig: go.Figure,
              layer: str, color: str):
    """ Helperfunctie om de lijnen uit de geopackage te vinden en toe te voegen aan de map. Layer is de naam van de
    laag in de geopackage waar de lijn is opgeslagen. Color is de kleur van deze lijn in de map. """

    gdf_traject = gpd.read_file(
        geoprob_pipe.input_data.app_settings.geopackage_filepath,
        layer=layer)
    gdf_traject = gdf_traject.to_crs("EPSG:4326")

    def plot_linestring(ls, display):
        xs, ys = ls.xy
        xs = list(xs)
        ys = list(ys)
        fig.add_trace(go.Scattermap(
            lon=xs,
            lat=ys,
            mode="lines",
            line=dict(color=color, width=1),
            hoverinfo="none",
            name=layer,
            legendgroup=layer,
            showlegend=display
        ))

    show = True
    for geom in gdf_traject.geometry:
        if isinstance(geom, LineString):
            plot_linestring(geom, show)
            show = False

        elif isinstance(geom, MultiLineString):
            for line in geom.geoms:
                plot_linestring(line, show)
                show = False

        elif isinstance(geom, GeometryCollection):
            for g in geom.geoms:
                if isinstance(g, LineString):
                    plot_linestring(g, show)
                    show = False
                elif isinstance(g, MultiLineString):
                    for line in g.geoms:
                        plot_linestring(line, show)
                        show = False

        else:
            print("Skipping unsupported geometry:", geom.geom_type)

    return fig


class BetaMap:

    def __init__(self, geoprob_pipe: GeoProbPipe, export: bool = False):

        self.geoprob_pipe = geoprob_pipe
        self.export = export

        # Logic
        self._import_results()
        self._setup_gdf()
        self._determine_zoom()
        self._create_figure()
        self._add_lines()
        self._optionally_export()

    def _import_results(self):
        # results import
        self.inp_point = self.geoprob_pipe.input_data.uittredepunten.gdf
        self.res_sc = self.geoprob_pipe.results.df_beta_scenarios_final
        mask = self.inp_point["uittredepunt_id"].isin(self.res_sc["uittredepunt_id"])
        self.inp_point = self.inp_point[mask]

        # Setup of beta category limits
        self.cg = (self.geoprob_pipe.input_data.traject_normering
                   .riskeer_categorie_grenzen)
        self.colors = ["rgb(30,141,41)", "rgb(146,206,90)",
                       "rgb(198,226,176)", "rgb(255,255,0)",
                       "rgb(254,165,3)", "rgb(255,0,0)",
                       "rgb(177,33,38)"]
        self.labels = ["+III", "+II", "+I", "0", "-I", "-II", "-III"]

        self.len_cg = self.cg[self.labels[0]][1] - self.cg[self.labels[6]][0]

        self.color1 = ((self.cg[self.labels[0]][1]
                        - self.cg[self.labels[6]][0]) / self.len_cg)
        self.color2 = ((self.cg[self.labels[1]][1]
                        - self.cg[self.labels[6]][0]) / self.len_cg)
        self.color3 = ((self.cg[self.labels[2]][1]
                        - self.cg[self.labels[6]][0]) / self.len_cg)
        self.color4 = ((self.cg[self.labels[3]][1]
                        - self.cg[self.labels[6]][0]) / self.len_cg)
        self.color5 = ((self.cg[self.labels[4]][1]
                        - self.cg[self.labels[6]][0]) / self.len_cg)
        self.color6 = ((self.cg[self.labels[5]][1]
                        - self.cg[self.labels[6]][0]) / self.len_cg)
        self.color7 = ((self.cg[self.labels[6]][1]
                        - self.cg[self.labels[6]][0]) / self.len_cg)
        self.color8 = ((self.cg[self.labels[6]][0]
                        - self.cg[self.labels[6]][0]) / self.len_cg)

    def _setup_gdf(self):
        self.hoverdata = ["uittredepunt_id", "converged", "beta"]

        self.df = self.res_sc.merge(self.inp_point, on="uittredepunt_id",
                                    how="inner")
        idx = self.df.groupby(["uittredepunt_id"])["beta"].idxmin()
        self.df = self.df.loc[idx]

        self.gdf = gpd.GeoDataFrame(
            self.df, geometry=gpd.points_from_xy(
                self.inp_point.geometry.x, self.inp_point.geometry.y
                ),
            crs="EPSG:28992")
        # Transformeer naar WGS84 (latitude / longitude)
        self.gdf_latlon = self.gdf.to_crs("EPSG:4326")

    def _determine_zoom(self):
        self.center_lat = self.gdf_latlon.geometry.y.mean()
        self.center_lon = self.gdf_latlon.geometry.x.mean()
        self.min_lat = self.gdf_latlon.geometry.y.min()
        self.max_lat = self.gdf_latlon.geometry.y.max()
        self.min_lon = self.gdf_latlon.geometry.x.min()
        self.max_lon = self.gdf_latlon.geometry.x.max()

        def _calculate_zoom(lat_range, lon_range):
            max_range = max(lat_range, lon_range)
            if max_range < 0.01:
                return 15
            elif max_range < 0.05:
                return 13
            elif max_range < 0.1:
                return 12
            elif max_range < 0.5:
                return 10
            elif max_range < 1.0:
                return 9
            else:
                return 8

        self.lat_range = self.max_lat - self.min_lat
        self.lon_range = self.max_lon - self.min_lon
        self.zoom = _calculate_zoom(self.lat_range, self.lon_range)

    def _create_figure(self):
        self.fig = go.Figure()
        self.fig.add_trace(go.Scattermap(
            mode="markers",
            lat=self.gdf_latlon.geometry.y,
            lon=self.gdf_latlon.geometry.x,
            marker=dict(
                size=9,
                color="black"
                ),
            showlegend=False
            ))
        self.fig.add_trace(go.Scattermap(
            mode='markers',
            lat=self.gdf_latlon.geometry.y,   # direct uit geometrie
            lon=self.gdf_latlon.geometry.x,   # direct uit geometrie
            marker=dict(
                size=8,
                color=self.gdf_latlon['beta'],
                colorscale=[
                    (self.color8, self.colors[6]),
                    (self.color7, self.colors[6]),
                    (self.color7, self.colors[5]),
                    (self.color6, self.colors[5]),
                    (self.color6, self.colors[4]),
                    (self.color5, self.colors[4]),
                    (self.color5, self.colors[3]),
                    (self.color4, self.colors[3]),
                    (self.color4, self.colors[2]),
                    (self.color3, self.colors[2]),
                    (self.color3, self.colors[1]),
                    (self.color2, self.colors[1]),
                    (self.color2, self.colors[0]),
                    (self.color1, self.colors[0])
                ],
                cmin=self.cg[self.labels[6]][0],
                cmax=self.cg[self.labels[0]][1],
                colorbar=dict(
                    title="Bèta, WBI cat.",
                    tickvals=[
                        self.cg[self.labels[6]][0],
                        self.cg[self.labels[6]][1],
                        self.cg[self.labels[5]][1],
                        self.cg[self.labels[4]][1],
                        self.cg[self.labels[3]][1],
                        self.cg[self.labels[2]][1],
                        self.cg[self.labels[1]][1],
                        self.cg[self.labels[0]][1]
                    ],
                    ticktext=[f"{v:.2f}" for v in [
                        self.cg[self.labels[6]][0],
                        self.cg[self.labels[6]][1],
                        self.cg[self.labels[5]][1],
                        self.cg[self.labels[4]][1],
                        self.cg[self.labels[3]][1],
                        self.cg[self.labels[2]][1],
                        self.cg[self.labels[1]][1],
                        self.cg[self.labels[0]][1]]],
                    )
                ),
            hoverinfo='text',
            text=self.gdf_latlon[self.hoverdata].apply(
                lambda row: '<br>'.join(
                    [f"{col}: {row[col]}" for col in self.hoverdata]
                    ),
                axis=1),
            showlegend=False
        ))

        # Layout
        self.fig.update_layout(
            map_style="open-street-map",
            # carto-positron, open-street-map, satellite-streets
            map_zoom=self.zoom,
            map_center=dict(
                lat=self.gdf_latlon.geometry.y.mean(),
                lon=self.gdf_latlon.geometry.x.mean()
            ),
            legend=dict(
                orientation="h",
                yanchor="bottom",
                y=1.02,
                xanchor="right",
                x=1
            ),
            dragmode='zoom',
            title='Faalkansberekening STPH'
        )

    def _add_lines(self):
        self.fig = _add_line(self.geoprob_pipe, self.fig,
                             "dijktraject", "black")
        self.fig = _add_line(self.geoprob_pipe, self.fig,
                             "intredelijn", "blue")
        self.fig = _add_line(self.geoprob_pipe, self.fig,
                             "binnenteenlijn", "purple")
        self.fig = _add_line(self.geoprob_pipe, self.fig,
                             "buitenteenlijn", "red")

    def _optionally_export(self):
        if self.export:
            path = self.geoprob_pipe.visualizations.maps.export_dir
            self.fig.write_html(os.path.join(
                path, 'Faalkansberekening STPH.html'), include_plotlyjs='cdn')
            # self.fig.write_image(os.path.join(
            #     path, 'Faalkansberekening STPH.png'), format='png')



import math
from typing import Optional
from scipy.stats import lognorm, norm


def calc_kar_waarde_lognormal(
        mean: float, percentiel: float, sd: Optional[float] = None, vc: Optional[float] = None, shift: float = 0.0
) -> float:
    r""" Berekening van de percentiel waarde van een lognormale verdeling met standaarddeviatie/variatie coëfficient
    en verschuiving.

    # TODO: Bron naar documentatie van deze methodiek.

    Args:
       mean (float): verwachtingswaarde van de lognormale verdeling
       sd (float): standaarddeviatie van de lognormale verdeling, verplicht indien vc niet gegeven
       vc (float): variatie coëfficient van de lognormale verdeling, verplicht indien sd niet gegeven
       percentiel (float): percentiel waarvoor de waarde van de lognormale verdeling wordt berekend. Waarde tussen 0.0 en 1.0.
       shift (float, optional): verschuiving. Defaults to 0.0.

    Returns:
       float: percentiel waarde van de lognormale verdeling
    """

    # Format input arguments
    if sd is None and vc is None:
        raise ValueError(
            f"Provide either the standard deviation (input argument sd), or the variation coefficient (vc).")
    if sd is None:
        sd = vc * mean

    # Logic
    vc_shift = float(sd / (mean - shift))
    log_sd = math.sqrt(math.log(1.0 + math.pow(vc_shift, 2.0)))
    log_mu = math.log(mean - shift) - 0.5 * math.pow(log_sd, 2.0)
    kar_waarde = float(lognorm.ppf(percentiel, s=log_sd, loc=0.0, scale=math.exp(log_mu))) + shift

    return kar_waarde


def calc_kar_waarde_normal(mean: float, std: float, percentiel: float) -> float:
    """Berekening van de percentiel waarde van een normale verdeling

    Args:
        mean (float): verwachtingswaarde van de normale verdeling
        std (float): standaarddeviatie van de normale verdeling
        percentiel (float): percentiel waarvoor de waarde van de normale verdeling wordt berekend

    Returns:
        float: percentiel waarde van de normale verdeling
    """
    return float(norm.ppf(percentiel, loc= mean, scale=std))


def convert_failure_probability_to_beta(failure_probability: float) -> float:
    """Converts failure probability (Pf) to the reliability index (β).

    The reliability index is the negative inverse of the standard normal
    cumulative distribution function (Φ) applied to the failure probability.

    β = -1 * Φ⁻¹(Pf)

    Args:
        failure_probability (float): failure probability (Pf)

    Returns:
        float: reliability index (β)
    """
    return float(-1 * norm.ppf(failure_probability))



from geoprob_pipe.calculations.systems.model4a.limit_state_functions import (
    limit_state_model4a, calc_Z_h, calc_Z_p, calc_Z_u, calc_Z_project)
from geoprob_pipe.calculations.systems.base_objects.system_calculation import SystemCalculation
from typing import List, Dict, Union, Tuple


class Model4aCalculation(SystemCalculation):
    """ Pre-defined system reliability calculation for Piping.

    Usage by calling the object, inserting the variable distributions and
    after that calling the run-method. The separate limit state models
    (uplift, heave and piping) are already pre-defined in the class.
    """

    def __init__(
            self,
            distributions: List[Dict],
            correlations: List[Tuple[str, str, float]] = None,
            reliability_settings: Dict[str, Union[str, float, int]] = None
    ):

        super().__init__(
            distributions=distributions,
            correlations=correlations,
            reliability_settings=reliability_settings)
        self.setup.variables_function = limit_state_model4a
        self.setup.system_limit_states = [calc_Z_u, calc_Z_h, calc_Z_p]
        self.setup.project_limit_state = calc_Z_project



from __future__ import annotations
import os
import sqlite3
import shutil
from openpyxl import load_workbook
from pandas import ExcelWriter, DataFrame
from geopandas import read_file, GeoDataFrame
from datetime import datetime
import importlib.resources
from typing import TYPE_CHECKING
from geoprob_pipe.calculations.systems.mappers.initial_input_mapper import INITIAL_INPUT_MAPPER
from geoprob_pipe.utils.validation_messages import BColors
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings
    from geoprob_pipe.cmd_app.parameter_input.input_parameter_tables import InputParameterTables


def join_vak_naam(df: DataFrame, app_settings: ApplicationSettings, ) -> DataFrame:
    gdf_vakindeling: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="vakindeling")
    gdf_vakindeling = gdf_vakindeling.rename(columns={"naam": "vak_naam", "id": "vak_id"})
    df = df.merge(gdf_vakindeling[["vak_naam", "vak_id"]], left_on="vak_id", right_on="vak_id")
    return df[["vak_id", "vak_naam", "naam", "kans"]]


def export_input_parameter_tables(app_settings: ApplicationSettings, tables: InputParameterTables):

    # Copy template to workspace
    dst_dir = os.path.join(
        app_settings.workspace_dir,
        "exports",
        app_settings.datetime_stamp,
        "parameter_input_process")
    os.makedirs(dst_dir, exist_ok=True)
    dst_path = os.path.join(dst_dir, "input_parameters_template.xlsx")
    dst_path = dst_path.replace("_template.xlsx", f".xlsx")
    if os.path.exists(dst_path):
        datetime_stamp = datetime.now().strftime("%Y-%m-%d_%H%M%S")
        dst_path = dst_path.replace(".xlsx", f"_{datetime_stamp}.xlsx")
    with importlib.resources.path(
            package='geoprob_pipe.cmd_app.parameter_input',
            resource='parameter_input_template.xlsx'
    ) as src_path:
        shutil.copy2(src=src_path, dst=dst_path)

    # Fill 'Model parameters'
    conn = sqlite3.connect(app_settings.geopackage_filepath)
    cursor = conn.cursor()
    cursor.execute("""
        SELECT geoprob_pipe_metadata."values" 
        FROM geoprob_pipe_metadata 
        WHERE metadata_type='geohydrologisch_model';
    """)
    result = cursor.fetchone()
    if not result:
        raise ValueError
    model_string = result[0]
    conn.close()
    df_dummy_data = DataFrame(INITIAL_INPUT_MAPPER[model_string]['input'])
    df_dummy_data = df_dummy_data.sort_values(by=["name"])
    df_model_parameters = df_dummy_data[["name", "description", "remark", "unit"]].copy()
    with ExcelWriter(dst_path, engine="openpyxl", mode="a", if_sheet_exists="overlay") as writer:
        df_model_parameters.to_excel(
            writer, sheet_name="Geohydrologisch model", index=False, header=False, startrow=6, startcol=0)
    # TODO: Kolom 'invullen' is nog onduidelijk en niet gevuld.
    # Geohydrologisch model in titel
    wb = load_workbook(dst_path)
    ws = wb["Geohydrologisch model"]
    ws["A1"] = f"Geohydrologisch model '{INITIAL_INPUT_MAPPER[model_string]['label']}'"
    wb.save(dst_path)

    # Fill 'Scenario invoer'
    df_scenario_invoer = tables.df_scenario_invoer
    df_scenario_invoer = join_vak_naam(df=df_scenario_invoer, app_settings=app_settings)
    with ExcelWriter(dst_path, engine="openpyxl", mode="a", if_sheet_exists="overlay") as writer:
        df_scenario_invoer.to_excel(
            writer, sheet_name="Scenario invoer", index=False, header=False, startrow=3, startcol=0)

    # Fill 'Parameter invoer'
    df_parameter_invoer = tables.df_parameter_invoer
    df_parameter_invoer = df_parameter_invoer[[
        "parameter", "scope", "scope_referentie", "ondergrondscenario_naam", "distribution_type", "mean", "variation",
        "deviation", "minimum", "maximum", "fragility_values_ref", "bronnen", "opmerking"]].copy()
    with ExcelWriter(dst_path, engine="openpyxl", mode="a", if_sheet_exists="overlay") as writer:
        df_parameter_invoer.to_excel(
            writer, sheet_name="Parameter invoer", index=False, header=False, startrow=4, startcol=0)

    # Fill 'Fragility values'
    df_fragility_values_invoer = tables.df_fragility_values_invoer
    df_fragility_values_invoer = df_fragility_values_invoer[["fragility_values_ref", "waarde", "kans"]].copy()
    with ExcelWriter(dst_path, engine="openpyxl", mode="a", if_sheet_exists="overlay") as writer:
        df_fragility_values_invoer.to_excel(
            writer, sheet_name="Fragility values", index=False, header=False, startrow=4, startcol=0)

    # Fill 'Correlatie invoer'
    df_correlatie_invoer = tables.df_correlatie_invoer
    df_correlatie_invoer = df_correlatie_invoer[["parameter_a", "parameter_b", "correlation"]].copy()
    with ExcelWriter(dst_path, engine="openpyxl", mode="a", if_sheet_exists="overlay") as writer:
        df_correlatie_invoer.to_excel(
            writer, sheet_name="Correlatie invoer", index=False, header=False, startrow=4, startcol=0)

    print(f"{BColors.UNDERLINE}Exporteren van invoer tabellen compleet:\n"
          f"{dst_path}{BColors.ENDC}")



from __future__ import annotations
from geoprob_pipe.cmd_app.spatial_layers.ahn import added_ahn, request_ahn
from typing import TYPE_CHECKING, Optional, Tuple, List
from geoprob_pipe.cmd_app.utils.spatial import load_dijktraject_linestring, load_hydra_nl_as_multipoint
from shapely import LineString, MultiPoint, Polygon, MultiPolygon, Point, unary_union
from shapely.geometry import mapping
import numpy as np
import rasterio.mask
import time
import random
from geopandas import GeoDataFrame
from copy import deepcopy
from geoprob_pipe.utils.validation_messages import BColors
from InquirerPy import inquirer

if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def request_buffer_distance() -> int:
    int_distance = 50
    distance_is_valid = False
    while distance_is_valid is False:
        str_distance: str = inquirer.text(
            message="Specificeer de afstand (in meters) waarbinnen gezocht moet worden aan binnendijkse zijde.\n"
                    "Zorg er voor dat deze afstand volledig binnen het toegevoegde raster valt. ",
        ).execute()

        # Validate
        try: int_distance = int(str_distance)
        except ValueError:
            print(f"{BColors.WARNING}Het lukte niet om de invoer om te zetten naar een integer (geheel getal). Weet je "
                  f"zeker dat de invoer correct is? Probeer het opnieuw.{BColors.ENDC}")
            continue
        if int_distance < 50:
            print(f"{BColors.WARNING}De opgegeven afstand moet ten minste 50 meter zijn. "
                  f"Vul opnieuw een waarde in. {BColors.ENDC}")
            continue

        # Valid
        distance_is_valid = True
    return int_distance


def create_buffer_binnendijks(
    app_settings: ApplicationSettings,
) -> Polygon:

    buffer_distance = request_buffer_distance()

    # Load data from geopackage
    ls_dijktraject: LineString = load_dijktraject_linestring(app_settings=app_settings)
    mp_hydra_nl_locaties: MultiPoint = load_hydra_nl_as_multipoint(app_settings=app_settings)

    # Create helper buffers
    buffer_both_sides = ls_dijktraject.buffer(distance=buffer_distance, cap_style='flat')
    buffer_single_side = ls_dijktraject.buffer(distance=buffer_distance, single_sided=True)
    buffer_other_side = buffer_both_sides.difference(buffer_single_side)

    # Determine which side is binnendijks
    buffer_binnendijks = buffer_single_side
    if mp_hydra_nl_locaties.intersects(buffer_binnendijks):
        buffer_binnendijks = buffer_other_side

    return buffer_binnendijks


def random_point_in_polygon(polygon):
    if isinstance(polygon, MultiPolygon):
        polygon = max(polygon.geoms, key=lambda poly: poly.area)

    minx, miny, maxx, maxy = polygon.bounds
    while True:
        p = Point(random.uniform(minx, maxx), random.uniform(miny, maxy))
        if polygon.contains(p):
            return p


def walk(walk_id: int, src, radius: float, pnt_walk_start: Point) -> Tuple[List, List, List, List]:
    iteration_id = 0
    distance = 999
    new_rows_circle_mid = []
    new_rows_circle = []
    new_rows_circle_lowest = []
    pnt_lowest: Optional[Point] = None
    while distance > 0.5 and iteration_id <= 99:
        iteration_id += 1

        pnt_center = pnt_lowest
        if pnt_center is None:
            pnt_center = pnt_walk_start
        circle = pnt_center.buffer(distance=radius)
        circle_geojson = [mapping(circle)]

        out_image, out_transform = rasterio.mask.mask(src, circle_geojson, crop=True)
        out_image = out_image[0]  # take first band
        out_image = np.where(out_image == src.nodata, np.nan, out_image)

        if np.all(np.isnan(out_image)):
            return [], [], [], [circle]

        min_val = np.nanmin(out_image)
        min_idx = np.nanargmin(out_image)
        row, col = np.unravel_index(min_idx, out_image.shape)
        rdx, rdy = rasterio.transform.xy(out_transform, int(row), col)
        pnt_lowest = Point(rdx, rdy)
        distance = pnt_center.distance(pnt_lowest)

        new_rows_circle_mid.append({"walk_id": walk_id, "iteration_id": iteration_id, "geometry": pnt_center})
        new_rows_circle.append({"walk_id": walk_id, "iteration_id": iteration_id, "geometry": circle})
        final_iteration = False
        if distance <= 1:
            final_iteration = True
        new_rows_circle_lowest.append(
            {"walk_id": walk_id, "iteration_id": iteration_id, "min_val": round(min_val, 2),
             "geometry": pnt_lowest, "final_iteration": final_iteration})

    circles_searched = [item['geometry'] for item in new_rows_circle]

    return new_rows_circle_mid, new_rows_circle, new_rows_circle_lowest, circles_searched


def search_lowes(app_settings: ApplicationSettings, buffer_binnendijks: Polygon, radius: float = 15.0):
    to_search_buffer = deepcopy(buffer_binnendijks)  #
    with rasterio.open(app_settings.ahn_filepath) as src:
        rows_circle_mid = []
        rows_circle = []
        rows_circle_lowest = []

        start_time = time.time()
        elapsed_time = 0.0
        interval_time = 2.0  # seconds
        next_report_time = elapsed_time + interval_time

        walk_id = 0
        while to_search_buffer.area / buffer_binnendijks.area > 0.05 and walk_id <= 9999:

            walk_id += 1

            # Get random point of largest polygon of multipolygon
            pnt_walk_start = random_point_in_polygon(to_search_buffer)

            # Walk
            new_rows_circle_mid, new_rows_circle, new_rows_circle_lowest, new_circles_searched = walk(
                walk_id=walk_id, src=src, pnt_walk_start=pnt_walk_start, radius=radius)

            # Subtract from buffer
            to_search_buffer = to_search_buffer.difference(unary_union(new_circles_searched))

            rows_circle_mid.extend(new_rows_circle_mid)
            rows_circle.extend(new_rows_circle)
            rows_circle_lowest.extend(new_rows_circle_lowest)

            end_time = time.time()
            elapsed_time = end_time - start_time
            if elapsed_time > next_report_time:
                next_report_time = next_report_time + interval_time
                progress_in_percentage = 100 - round((to_search_buffer.area / buffer_binnendijks.area)*100, 0)
                print(f"Elapsed time walking: {elapsed_time:.2f} seconds. "
                      f"Current walk number: {walk_id}. "
                      f"Searched buffer for {progress_in_percentage}% of buffer")

    progress_in_percentage = 100 - round((to_search_buffer.area / buffer_binnendijks.area) * 100, 0)
    print(f"Completed walking buffer in {elapsed_time:.2f} seconds. "
          f"Total of {walk_id} walks. "
          f"Covered {progress_in_percentage}% of buffer. "
          f"Remainder is considered irrelevant.")

    gdf_circle_lowests = GeoDataFrame(rows_circle_lowest, crs='EPSG:28992')
    gdf_proposed_uittredepunten: GeoDataFrame = gdf_circle_lowests[gdf_circle_lowests['final_iteration'] == True]

    # Filter duplicates
    indices_to_keep = []
    for row in gdf_proposed_uittredepunten.itertuples():
        # noinspection PyUnresolvedReferences
        pnt = row.geometry.buffer(1)
        indices = gdf_proposed_uittredepunten[gdf_proposed_uittredepunten.geometry.intersects(pnt)].index.values
        indices.sort()
        indices_to_keep.append(indices[0])
    gdf_proposed_uittredepunten = gdf_proposed_uittredepunten[gdf_proposed_uittredepunten.index.isin(indices_to_keep)]

    gdf_proposed_uittredepunten: GeoDataFrame = gdf_proposed_uittredepunten[["walk_id", "min_val", "geometry"]]
    print(f"Found {gdf_proposed_uittredepunten.__len__()} proposed exit points.")

    return gdf_proposed_uittredepunten


def algorithm_walking_circles(app_settings: ApplicationSettings):

    if not added_ahn(app_settings=app_settings, display_added_msg=False):
        request_ahn(app_settings=app_settings)

    buffer_binnendijks = create_buffer_binnendijks(app_settings=app_settings)
    gdf_proposed_uittredepunten = search_lowes(app_settings=app_settings, buffer_binnendijks=buffer_binnendijks)

    gdf_to_add = gdf_proposed_uittredepunten[["geometry", "min_val"]]
    gdf_to_add = gdf_to_add.rename(columns={"min_val": "mv_exit"})
    gdf_to_add.to_file(app_settings.geopackage_filepath, layer="uittredepunten", driver="GPKG")
    print(BColors.OKBLUE, f"✅  Uittredepunten toegevoegd.", BColors.ENDC)



from __future__ import annotations
from InquirerPy import inquirer
import sqlite3
from geoprob_pipe.cmd_app.parameter_input.expand_input_tables import run_expand_input_tables
from geoprob_pipe.cmd_app.parameter_input.initiate_input_excel_tables import initiate_input_excel_tables
from geoprob_pipe.cmd_app.parameter_input.input_parameter_figures import InputParameterFigures
from geoprob_pipe.cmd_app.parameter_input.export_input_parameter_excel import export_input_parameter_tables
from geoprob_pipe.cmd_app.parameter_input.input_parameter_tables import InputParameterTables
from typing import TYPE_CHECKING, Optional
import os
import sys
from geoprob_pipe.utils.validation_messages import BColors
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def possibly_initiatie_input_tables_in_db(app_settings: ApplicationSettings):

    # Get table names
    conn = sqlite3.connect(app_settings.geopackage_filepath)
    cursor = conn.cursor()
    cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
    tables_names = [row[0] for row in cursor.fetchall()]
    conn.close()

    # If already exist
    if ("parameter_invoer" in tables_names and
            "scenario_invoer" in tables_names and "fragility_values_invoer" in tables_names):
        # print(f"{BColors.UNDERLINE}Tables already exist in geopackage{BColors.ENDC}")
        return

    # If it doesn't exist yet
    # print(f"{BColors.UNDERLINE}Tables do not yet exist in geopackage{BColors.ENDC}")
    initiate_input_excel_tables(app_settings=app_settings)
    print(f"{BColors.UNDERLINE}Basis invoer tabellen zijn nu geïnitieerd in het GeoProb-Pipe-bestand. Deze kun je naar "
          f"wens aanpassen in het vervolgproces.{BColors.ENDC}")


def load_tables_from_db(app_settings: ApplicationSettings) -> InputParameterTables:
    tables = InputParameterTables(geopackage_filepath=app_settings.geopackage_filepath)
    return tables


def validate_raw_input_tables(
        app_settings: ApplicationSettings, tables: InputParameterTables
) -> bool:
    return tables.validate_and_report(app_settings=app_settings)


def inquire_if_input_figures_should_be_exported(app_settings: ApplicationSettings, tables: InputParameterTables):
    choices_list = [
        "Ja",
        "Nee",
        "Applicatie afsluiten"
    ]
    choice = inquirer.select(
        message="Wil je overzichtsfiguren exporteren van de invoer tabellen? \n"
                "Hierin kun je gemakkelijk zien voor welke parameter welke invoer is ingesteld, en op welk niveau "
                "(geografisch-, traject-, vak-, of uittredepunt-niveau).",
        choices=choices_list,
        default=choices_list[0],
    ).execute()

    if choice == choices_list[0]:
        InputParameterFigures(app_settings=app_settings, tables=tables, export=True)
    elif choice == choices_list[1]:
        pass  # Just continue
    elif choice == choices_list[2]:
        sys.exit(f"Applicatie is afgesloten.")
    else:
        raise ValueError


def validate_expanded_input_tables(app_settings: ApplicationSettings) -> bool:
    df_expanded = run_expand_input_tables(geopackage_filepath=app_settings.geopackage_filepath)
    df_nans = df_expanded[df_expanded['parameter_input'].isna()]

    # No issues?
    if df_nans.__len__() == 0:
        return True

    # Report issues back
    export_dir = os.path.join(
        os.path.dirname(app_settings.geopackage_filepath),
        "exports",
        str(app_settings.datetime_stamp),
        "parameter_input_process")
    os.makedirs(export_dir, exist_ok=True)
    export_path = os.path.join(export_dir, "validation_missing_parameter_input.xlsx")
    if os.path.exists(export_path):
        os.remove(export_path)
    df_nans.to_excel(export_path)
    print(f"{BColors.WARNING}Er mist parameter invoer voor {df_nans.__len__()} berekeningen.\n"
          f"Dit is voor {df_nans['parameter_name'].unique().__len__()} unieke parameters, "
          f"{df_nans['uittredepunt_id'].unique().__len__()} unieke uittredepunten en "
          f"{df_nans['ondergrondscenario_naam'].unique().__len__()} unieke ondergrondscenarios.\n"
          f"De gedetailleerde lijst is geëxporteerd naar\n"
          f"{export_path}{BColors.ENDC}")
    return False


def inquire_to_import_export_tables_and_figures_or_continue(
        app_settings: ApplicationSettings, tables: InputParameterTables,
        validity_raw_tables: bool, validity_extended_tables: bool
):

    # Determine options
    choices_list = []
    if validity_raw_tables and validity_extended_tables:
        choices_list.append("Zijn de invoer tabellen zijn naar wens? Ga door naar volgende stap")
    else:
        choices_list.append(
            "Zijn de invoer tabellen zijn naar wens? Ga door naar volgende stap (n.v.t. -> invoer niet valide)")
    if validity_raw_tables:
        choices_list.append("Overzichtsfiguren van invoertabellen: Exporteren")
    choices_list.extend([
        "Invoer tabellen: Importeren vanuit Excel",
        "Invoer tabellen: Exporteren naar Excel",
        "Toelichting per keuze optie",
        "Applicatie afsluiten"])

    # Provide user options
    choice = inquirer.select(
        message="Maak een keuze voor het gereedmaken van de invoertabellen.",
        choices=choices_list,
        default=choices_list[0],
    ).execute()

    if choice == "Zijn de invoer tabellen zijn naar wens? Ga door naar volgende stap":
        # run_calculations(geopackage_filepath=app_settings.geopackage_filepath)
        print(BColors.OKBLUE, f"✔  Parameter invoer afgerond.", BColors.ENDC)
        return

    elif choice == "Zijn de invoer tabellen zijn naar wens? Ga door naar volgende stap (n.v.t. -> invoer niet valide)":
        inquire_to_import_export_tables_and_figures_or_continue(
            app_settings=app_settings, tables=tables, validity_raw_tables=validity_raw_tables,
            validity_extended_tables=validity_extended_tables)

    elif choice == "Overzichtsfiguren van invoertabellen: Exporteren":
        InputParameterFigures(app_settings=app_settings, tables=tables, export=True)
        inquire_to_import_export_tables_and_figures_or_continue(
            app_settings=app_settings, tables=tables, validity_raw_tables=validity_raw_tables,
            validity_extended_tables=validity_extended_tables)

    elif choice == "Invoer tabellen: Importeren vanuit Excel":
        process_import_input(app_settings=app_settings)

    elif choice == "Invoer tabellen: Exporteren naar Excel":
        process_export_input_of_db(
            app_settings=app_settings, tables=tables, validity_raw_tables=validity_raw_tables,
            validity_extended_tables=validity_extended_tables)

    elif choice == "Toelichting per keuze optie":
        print("""
Invoer tabellen zijn naar wens, ga door naar volgende stap  ->  Indien je deze keuzemogelijkheid krijgt zijn de invoertabellen valide en kun je door naar de volgende stap.
                                                                Je zegt daarmee eveneens dat de invoertabellen naar wens zijn. 
Overzichtsfiguren van invoertabellen: Exporteren            ->  Deze interactive HTML-figuren geven je per parameter een snel visueel overzicht van de invoer in het GeoProb-Pipe-bestand. 
Invoer tabellen: Importeren vanuit Excel                    ->  Importeer vanuit Excel de invoertabellen om ze te laten valideren, visualiseren en/of op te slaan in het GeoProb-Pipe-bestand. 
Invoer tabellen: Exporteren naar Excel                      ->  Exporteer vanuit het GeoProb-Pipe-bestand de invoertabellen om ze in Excel te bekijken en/of verder aan te vullen. 
        """)
        inquire_to_import_export_tables_and_figures_or_continue(
            app_settings=app_settings, tables=tables, validity_raw_tables=validity_raw_tables,
            validity_extended_tables=validity_extended_tables)

    elif choice == "Applicatie afsluiten":
        sys.exit(f"Applicatie is afgesloten.")

    else:
        raise ValueError


def export_input_tables_of_db(app_settings: ApplicationSettings, tables: InputParameterTables):
    export_input_parameter_tables(app_settings=app_settings, tables=tables)


def import_input_tables(geopackage_filepath: str) -> InputParameterTables:

    filepath: Optional[str] = None
    filepath_is_valid = False
    while filepath_is_valid is False:
        filepath: str = inquirer.text(
            message="Specificeer het volledige bestandspad naar het input parameters Excel.",
        ).execute()

        filepath = filepath.replace('"', '')

        if not filepath.endswith(".xlsx"):
            print(BColors.WARNING, f"Het bestand moet een .xlsx-bestand zijn. Jouw invoer "
                                   f"{os.path.basename(filepath)} eindigt niet op deze extensie.", BColors.ENDC)
            continue
        if not os.path.exists(filepath):
            print(BColors.WARNING, f"Het opgegeven bestandspad bestaat niet.", BColors.ENDC)
            continue

        filepath_is_valid = True

    tables = InputParameterTables(path_to_excel=filepath, geopackage_filepath=geopackage_filepath)
    print(f"{BColors.UNDERLINE}Tabellen zijn nu geïmporteerd.{BColors.ENDC}")
    return tables


def inquire_to_store_input_tables_to_db(
        app_settings: ApplicationSettings, tables: InputParameterTables):

    choices_list = ["Ja", "Nee"]
    choice = inquirer.select(
        message="Wil je de nieuwe tabellen opslaan in het GeoProb-Pipe-bestand?",
        choices=choices_list,
        default=choices_list[0],
    ).execute()

    if choice == "Ja":
        conn = sqlite3.connect(app_settings.geopackage_filepath)
        tables.df_scenario_invoer.to_sql("scenario_invoer", conn, if_exists="replace", index=False)
        tables.df_parameter_invoer.to_sql("parameter_invoer", conn, if_exists="replace", index=False)
        tables.df_fragility_values_invoer.to_sql(
            "fragility_values_invoer", conn, if_exists="replace", index=False)
        tables.df_correlatie_invoer.to_sql("correlatie_invoer", conn, if_exists="replace", index=False)
        conn.close()
        print(f"{BColors.UNDERLINE}Tabellen zijn nu opgeslagen in het GeoProb-Pipe-file.{BColors.ENDC}")

    elif choice == "Nee":
        pass  # Just continue

    else:
        raise ValueError


def process_input_exist_in_db(app_settings: ApplicationSettings):
    possibly_initiatie_input_tables_in_db(app_settings=app_settings)
    tables: InputParameterTables = load_tables_from_db(app_settings=app_settings)

    # Validate raw tables
    validity_raw_tables = validate_raw_input_tables(app_settings=app_settings, tables=tables)

    # Validate expanded tables
    validity_extended_tables: Optional[bool] = None
    if validity_raw_tables: validity_extended_tables = validate_expanded_input_tables(app_settings=app_settings)

    # Provide user with follow-up options
    inquire_to_import_export_tables_and_figures_or_continue(
        app_settings=app_settings, tables=tables,
        validity_raw_tables=validity_raw_tables, validity_extended_tables=validity_extended_tables)
    # -> Redirects also to new process loop


def process_export_input_of_db(
        app_settings: ApplicationSettings, tables: InputParameterTables,
        validity_raw_tables: bool, validity_extended_tables: bool):
    export_input_tables_of_db(app_settings=app_settings, tables=tables)
    inquire_to_import_export_tables_and_figures_or_continue(
        app_settings=app_settings, tables=tables, validity_raw_tables=validity_raw_tables,
        validity_extended_tables=validity_extended_tables)
    # -> Redirects also to new process loop


def process_import_input(app_settings: ApplicationSettings):
    """ Questionnaire process for importing a new version of the input table Excel. """
    tables = import_input_tables(app_settings.geopackage_filepath)  # Asks user for path to input-file

    # Validate raw tables
    validity_raw_tables = validate_raw_input_tables(app_settings=app_settings, tables=tables)

    # Ask to export overview pictures
    if validity_raw_tables: inquire_if_input_figures_should_be_exported(app_settings=app_settings, tables=tables)

    # Validate expanded tables
    validity_extended_tables: Optional[bool] = None
    if validity_raw_tables: validity_extended_tables = validate_expanded_input_tables(app_settings=app_settings)

    # Provide user with follow-up options
    inquire_to_store_input_tables_to_db(app_settings=app_settings, tables=tables)
    inquire_to_import_export_tables_and_figures_or_continue(
        app_settings=app_settings, tables=tables, validity_raw_tables=validity_raw_tables,
        validity_extended_tables=validity_extended_tables)
    # -> Redirects also to new process loop


def added_input_parameter_data(app_settings: ApplicationSettings) -> bool:
    process_input_exist_in_db(app_settings=app_settings)
    return True



from probabilistic_library import DistributionType


INITIAL_INPUT = [
    {
        "name": "L_kwelweg",
        "description": "TODO",  # TODO
        "remark": "",
        "unit": "m",
        "distribution_type": DistributionType.deterministic,  # TODO
        "mean": 150.0,  # TODO
        "source": "spatial",  # TODO
    },
    {
        "name": "polderpeil",
        "description": "Polderpeil ter plaatse van uittredepunten.",
        "remark": "",
        "unit": "m+NAP",
        "distribution_type": DistributionType.deterministic,
        "mean": 0.0,
        "source": "spatial",
    },
    {
        "name": "buitenwaterstand",
        "description": "Buitenwaterstand. Indien keuze voor 'cdf_curve' als distributie type, dan wordt er een "
                       "een overschrijdingsfrequentielijn berekend met de Hydra-NL Python wrapper Pydra.",
        "remark": "",
        "unit": "m+NAP",
        "distribution_type": DistributionType.deterministic,
        "mean": 5.0,
        "source": "excel",
    },
    {
        "name": "mv_exit",
        "description": "Bodemhoogte ter plaatse van uittredepunten.",
        "remark": "",
        "unit": "m+NAP",
        "distribution_type": DistributionType.deterministic,
        "mean": 0.5,
        "source": "spatial",
    },
    {
        "name": "top_zand",
        "description": "Geschematiseerde top van het zand in het vak.",
        "remark": "",
        "unit": "m+NAP",
        "distribution_type": DistributionType.normal,
        "mean": -3.0,
        "deviation": 0.5,
        "source": "excel",
    },
    {
        "name": "r_exit",
        "description": "TODO",  # TODO
        "remark": "",
        "unit": "-",  # TODO
        "distribution_type": DistributionType.normal,  # TODO
        "mean": -0.5,  # TODO
        "deviation": 0.05,  # TODO
        "source": "excel",  # TODO
    },
    {
        "name": "k_wvp",
        "description": "",  # TODO
        "remark": "",  # TODO
        "unit": "",  # TODO
        "distribution_type": DistributionType.log_normal,  # TODO
        "mean": 2000.0,  # TODO
        "variation": 0.35,  # TODO
        "source": "excel",  # TODO
    },
    {
        "name": "modelfactor_h",
        "description": "Modelfactor voor heave.",
        "remark": "",
        "unit": "[-]",
        "distribution_type": DistributionType.log_normal,
        "mean": 1.0,
        "variation": 0.1,
        "source": "excel",
    },
    {
        "name": "modelfactor_ff",
        "description": "Modelfactor voor fijne fractie.",
        "remark": "",
        "unit": "[-]",
        "distribution_type": DistributionType.log_normal,  # TODO
        "mean": 1.0,  # TODO
        "variation": 0.1,  # TODO
        "source": "excel",  # TODO
    },
    {
        "name": "modelfactor_3d",
        "description": "Modelfactor voor 3D effecten.",
        "remark": "",
        "unit": "[-]",
        "distribution_type": DistributionType.log_normal,  # TODO
        "mean": 1.0,  # TODO
        "variation": 0.1,  # TODO
        "source": "excel",  # TODO
    },
    {
        "name": "modelfactor_aniso",
        "description": "Modelfactor voor anisotropie.",
        "remark": "",
        "unit": "[-]",
        "distribution_type": DistributionType.log_normal,  # TODO
        "mean": 1.0,  # TODO
        "variation": 0.1,  # TODO
        "source": "excel",  # TODO
    },
    {
        "name": "modelfactor_ml",
        "description": "Modelfactor voor meerlaagsheid zandpakket.",
        "remark": "",
        "unit": "[-]",
        "distribution_type": DistributionType.log_normal,  # TODO
        "mean": 1.0,  # TODO
        "variation": 0.1,  # TODO
        "source": "excel",  # TODO
    },
    {
        "name": "i_c_h",
        "description": "Kritieke heave gradiënt.",
        "remark": "",
        "unit": "[-]",
        "distribution_type": DistributionType.log_normal,
        "mean": 0.7,
        "variation": 0.15,
        "source": "excel",
    },
    {
        "name": "D_wvp",
        "description": "Dikte van het watervoerende pakket.",
        "remark": "",
        "unit": "m",
        "distribution_type": DistributionType.log_normal,
        "mean": 50,
        "deviation": 1.5,
        "source": "excel",
    },
    {
        "name": "modelfactor_u",
        "description": "Modelfactor voor uplift.",
        "remark": "",
        "unit": "[-]",
        "distribution_type": DistributionType.log_normal,
        "mean": 1.0,
        "variation": 0.1,
        "source": "excel",
    },
    {
        "name": "gamma_water",
        "description": "Volumegewicht van water.",
        "remark": "",
        "unit": "kN/m³",
        "distribution_type": DistributionType.deterministic,
        "mean": 9.81,
        "source": "excel",
    },
    {
        "name": "gamma_sat_deklaag",
        "description": "Gemiddeld volumegewicht van de deklaag.",
        "remark": "",
        "unit": "kN/m³",
        "distribution_type": DistributionType.log_normal,
        "mean": 13.9,
        "variation": 0.05,
        "source": "excel",
    },
    {
        "name": "modelfactor_p",
        "description": "Modelfactor voor piping.",
        "remark": "",
        "unit": "[-]",
        "distribution_type": DistributionType.log_normal,
        "mean": 1.0,
        "variation": 0.1,
        "source": "excel",
    },
    {
        "name": "d70",
        "description": "70% percentiel van de korrelverdeling.",
        "remark": "",
        "unit": "m",
        "distribution_type": DistributionType.log_normal,
        "mean": 2.8e-4,
        "variation": 0.15,
        "source": "excel",
    },
    {
        "name": "g",
        "description": "Zwaartekrachtversnelling.",
        "remark": "",
        "unit": "m/s²",
        "distribution_type": DistributionType.deterministic,
        "mean": 9.81,
        "source": "excel",
    },
    {
        "name": "v",
        "description": "Kinematische viscositeit – Sellmeijer.",
        "remark": "",
        "unit": "m²/s",
        "distribution_type": DistributionType.deterministic,
        "mean": 0.00000133,
        "source": "excel",
    },
    {
        "name": "theta",
        "description": "Rolweerstandshoek – Sellmeijer.",
        "remark": "",
        "unit": "graden",
        "distribution_type": DistributionType.deterministic,
        "mean": 37.0,
        "source": "excel",
    },
    {
        "name": "eta",
        "description": "White's weerstandscoëfficiënt (sleepkrachtfactor, constante van White) – Sellmeijer",
        "remark": "",
        "unit": "[-]",
        "distribution_type": DistributionType.deterministic,
        "mean": 0.25,
        "source": "excel",
    },
    {
        "name": "d70_m",
        "description": "Referentiewaarde voor de 70% percentiel van de korrelverdeling – Sellmeijer.",
        "remark": "",
        "unit": "m",
        "distribution_type": DistributionType.deterministic,
        "mean": 2.08e-4,
        "source": "excel",
    },
    {
        "name": "gamma_korrel",
        "description": "Volumieke dichtheid zand onder water – Sellmeijer.",
        "remark": "",
        "unit": "kN/m³",
        "distribution_type": DistributionType.deterministic,
        "mean": 16.5,
        "source": "excel",
    },
    {
        "name": "r_c_deklaag",
        "description": "Reductie constante van het verval over de deklaag.",
        "remark": "",
        "unit": "[-]",
        "distribution_type": DistributionType.deterministic,
        "mean": 0.3,
        "source": "excel",
    },
]



from __future__ import annotations
import sqlite3
from geoprob_pipe.calculations.systems.mappers.calculation_mapper import CALCULATION_MAPPER
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from geoprob_pipe.calculations.systems.base_objects.base_system_build import BaseSystemBuilder
    from geoprob_pipe.calculations.systems.base_objects.system_calculation import SystemCalculation


EXAMPLE_SCRIPT_REPRODUCING_SINGLE_CALCULATION = r"""
from geoprob_pipe import (
    reproduce_single_calculation, ParallelSystemReliabilityCalculation)

calc: ParallelSystemReliabilityCalculation = reproduce_single_calculation(
    geopackage_filepath=r"/pad/naar/het/bestand/geoprob_pipe.gpkg",
    uittredepunt_id=1234,            # Replace with id of interest
    ondergrondscenario_naam="PL",    # Replace with scenario name of interest
)
"""


EXPLANATION_REPRODUCING_SINGLE_CALCULATION = """
Met de onderstaande code kun je in de Python console een enkele berekening reproduceren en inspecteren. Dit is 
bijvoorbeeld handig wanneer je de Python objecten wilt vergelijken met de gegenereerde in- en uitvoer. Of wanneer je 
een vergelijk maakt met de PTK-tool.

Om dit te doen kopieer je de onderstaande code naar de Python console. Zorg er voor dat je de Python console gebruikt 
waarin je GeoProb-Pipe ook hebt geïnstalleerd. Vervang eveneens het uittredepunt id met het id wat je wilt bekijken. 
Hetzelfde doe je voor de ondergrondscenario_naam.

Het reproduceren en herberekenen duurt slechts enkele seconden. Daarna kun je de objecten inspecteren.

Let op: Deze feature vergt enige ervaring met Python.
"""


def reproduce_single_calculation(
        geopackage_filepath: str, uittredepunt_id: int, ondergrondscenario_naam: str,
) -> SystemCalculation:

    # Fetch geohydrological model
    conn = sqlite3.connect(geopackage_filepath)
    cursor = conn.cursor()
    cursor.execute(
        'SELECT "values" FROM geoprob_pipe_metadata WHERE metadata_type = ?',
        ("geohydrologisch_model",))
    geohydrologisch_model: str = cursor.fetchone()[0]

    # Fetch vak id
    cursor.execute(
        'SELECT vak_id FROM uittredepunten WHERE uittredepunt_id = ?',
        (uittredepunt_id,))
    vak_id: int = int(cursor.fetchone()[0])
    conn.close()

    # Construct calculation builder
    builder: BaseSystemBuilder = (
        CALCULATION_MAPPER[geohydrologisch_model]["system_builder"](
            geopackage_filepath=geopackage_filepath, to_run_vakken_ids=None))

    # Construct calculation
    row = {"uittredepunt_id": uittredepunt_id,
           "ondergrondscenario_naam": ondergrondscenario_naam,
           "vak_id": vak_id}
    calc = builder.build_instance(row_unique=row)
    calc.run()

    return calc



from __future__ import annotations
from geoprob_pipe.utils.statistics import convert_failure_probability_to_beta
from pandas import DataFrame, concat
import numpy as np
from typing import TYPE_CHECKING, List
if TYPE_CHECKING:
    from geoprob_pipe.results import Results
    from geoprob_pipe import GeoProbPipe
    from geoprob_pipe.calculations.systems.base_objects.system_calculation import \
        SystemCalculation
    from probabilistic_library import DesignPoint
    from geoprob_pipe.calculations.systems.build_and_run import CalcResult


def collect_df_beta_limit_state(calculation: SystemCalculation) -> DataFrame:

    def create_row(dp: DesignPoint, model_name):
        return {
            "uittredepunt_id": calculation.metadata["uittredepunt_id"],
            "ondergrondscenario_id": calculation.metadata["ondergrondscenario_naam"],  # TODO: id naar naam veranderen?
            "vak_id": calculation.metadata["vak_id"],
            "limit_state": model_name,
            "converged": dp.is_converged,
            "beta": round(dp.reliability_index, 2),
            "failure_probability": dp.probability_failure,
            "convergence": dp.convergence,
            "total_iterations": dp.total_iterations,
            "total_model_runs": dp.total_model_runs,
        }

    rows = []
    for design_point, model in zip(calculation.results.dps_limit_states, calculation.setup.system_limit_states):
        rows.append(create_row(dp=design_point, model_name=model.__name__))
    df = DataFrame(rows).sort_values(by=["uittredepunt_id", "ondergrondscenario_id", "vak_id"]).reset_index(drop=True)
    return df


def combine_df_beta_per_limit_state(calc_results: List[CalcResult]) -> DataFrame:
    assert calc_results.__len__() > 0
    df = concat((result.df_limit_state for result in calc_results), ignore_index=True)
    return df


def collect_df_beta_scenario_rp(calc: SystemCalculation) -> DataFrame:
    return DataFrame([{
        "uittredepunt_id": calc.metadata["uittredepunt_id"],
        "ondergrondscenario_id": calc.metadata["ondergrondscenario_naam"],  # TODO: id naar naam veranderen?
        "vak_id": calc.metadata["vak_id"],
        "system_calculation": calc,
        "converged": calc.results.dp_reliability.is_converged,
        "beta": round(calc.results.dp_reliability.reliability_index, 2),
        "failure_probability": calc.results.dp_reliability.probability_failure,
        "convergence": calc.results.dp_reliability.convergence,
        "total_model_runs": calc.results.dp_reliability.total_model_runs,
        "total_iterations": calc.results.dp_reliability.total_iterations,
    }])


def collect_df_beta_scenario_cp(calc: SystemCalculation) -> DataFrame:
    return DataFrame([{
        "uittredepunt_id": calc.metadata["uittredepunt_id"],
        "ondergrondscenario_id": calc.metadata["ondergrondscenario_naam"],  # TODO: id naar naam veranderen?
        "vak_id": calc.metadata["vak_id"],
        "system_calculation": calc,
        "converged": calc.results.dp_combine.is_converged,
        "beta": round(calc.results.dp_combine.reliability_index, 2),
        "failure_probability": calc.results.dp_combine.probability_failure,
        "convergence": calc.results.dp_combine.convergence,
        "total_model_runs": calc.results.dp_combine.total_model_runs,
    }])


def collect_df_beta_scenario_final(calc: SystemCalculation) -> DataFrame:
    """ Converts a SystemCalculation-object to a single-row DataFrame with the final result of the scenario
    calculations. Because there are several calculation methods, and the preferred result also depends on convergence,
    we use the below flow chart to determine the final result.

    .. image:: /_static/flow-chart-final-result-scenario-calculations.png
       :alt: Flow chart final result scenario calculations
       :align: center

    :param calc:
    :return:
    """

    # Base DataFrame-row to return
    beta1: float = calc.results.dp_combine.reliability_index
    converged1: bool = calc.results.dp_combine.is_converged
    beta2: float = calc.results.dp_reliability.reliability_index
    converged2: bool = calc.results.dp_reliability.is_converged
    beta3: float = max([dp.reliability_index for dp in calc.results.dps_limit_states])
    converged3: bool = all([dp.is_converged for dp in calc.results.dps_limit_states])
    return_dict = {
        "uittredepunt_id": calc.metadata["uittredepunt_id"],
        "ondergrondscenario_id": calc.metadata["ondergrondscenario_naam"],  # TODO: id naar naam veranderen?
        "vak_id": calc.metadata["vak_id"],
        "system_calculation": calc,
        "beta1": beta1, "converged1": converged1,
        "beta2": beta2, "converged2": converged2,
        "beta3": beta3, "converged3": converged3,
    }

    # Flow chart step 1: Combine Project converged?
    pof1: float = calc.results.dp_combine.probability_failure
    if converged1:
        return_dict["method_used"] = "1: Combine Project"
        return_dict["flow_chart_number"] = 1
        return_dict["failure_probability"] = pof1
        return_dict["beta"] = beta1
        return_dict["converged"] = converged1
        return_dict["advise"] = "-"
        return DataFrame([return_dict])

    # Flow chart step 2: Reliability Project converged?
    pof2: float = calc.results.dp_reliability.probability_failure
    if converged2:
        return_dict["method_used"] = "2: Reliability Project"
        return_dict["flow_chart_number"] = 2
        return_dict["failure_probability"] = pof2
        return_dict["beta"] = beta2
        return_dict["converged"] = converged2
        return_dict["advise"] = "-"
        return DataFrame([return_dict])

    # Flow chart step 3: Separate Limit States all converged?
    pof3: float = min([dp.probability_failure for dp in calc.results.dps_limit_states])
    if converged3:
        return_dict["method_used"] = "3: Max Limit States"
        return_dict["flow_chart_number"] = 3
        return_dict["failure_probability"] = pof3
        return_dict["beta"] = beta3
        return_dict["converged"] = converged3
        return_dict["advise"] = ("Result is probably a conservative approximation.You could consider to find "
                                 "convergence for the Combine or Reliability method.")
        return DataFrame([return_dict])

    # Flow chart step 4: B >= 8.0 (of all methods)?
    all_pofs: List[float] = [pof3, pof2, pof1]
    all_methods: List[str] = ["3: Max Limit States", "2: Reliability Project", "1: Combine Project"]
    all_betas: List[float] = [beta3, beta2, beta1]
    index_max_pof = np.argmax(all_pofs)
    if beta1 >= 8.0 and beta2 >= 8.0 and beta3 >= 8.0:
        return_dict["method_used"] = all_methods[index_max_pof]
        return_dict["flow_chart_number"] = 4
        return_dict["failure_probability"] = all_pofs[index_max_pof]
        return_dict["beta"] = all_betas[index_max_pof]
        return_dict["converged"] = False
        return_dict["advise"] = "Result is that positive, no reason to fine tune."
        return DataFrame([return_dict])

    # No positive result
    return_dict["method_used"] = all_methods[index_max_pof]
    return_dict["flow_chart_number"] = 5
    return_dict["failure_probability"] = all_pofs[index_max_pof]
    return_dict["beta"] = all_betas[index_max_pof]
    return_dict["converged"] = False
    return_dict["advise"] = "Consider fine tuning the calculation settings (of e.g. FORM, Importance Sampling, etc)."
    return DataFrame([return_dict])


def combine_df_beta_per_scenario_rp(calc_results: List[CalcResult]) -> DataFrame:
    df = concat((result.df_scenario_rp for result in calc_results), ignore_index=True)
    df = df.sort_values(["uittredepunt_id", "ondergrondscenario_id", "vak_id"]).reset_index(drop=True)
    return df


def combine_df_beta_per_scenario_cp(calc_results: List[CalcResult]) -> DataFrame:
    df = concat((result.df_scenario_cp for result in calc_results), ignore_index=True)
    df = df.sort_values(["uittredepunt_id", "ondergrondscenario_id", "vak_id"]).reset_index(drop=True)
    return df


def combine_df_beta_per_scenario_final(calc_results: List[CalcResult]) -> DataFrame:
    df = concat((result.df_scenario_final for result in calc_results), ignore_index=True)
    df = df.sort_values(["uittredepunt_id", "ondergrondscenario_id", "vak_id"]).reset_index(drop=True)
    return df


def calculate_df_beta_per_uittredepunt(geoprob_pipe: GeoProbPipe, results: Results) -> DataFrame:
    """ Generates the DataFrame of the final result for the exit points.

    Because there is an automated decision-making in the scenario calculations (see flow chart over there), for the exit
    points the flow chart is extended below.

    .. image:: /_static/flow-chart-final-result-uittredepunt-calculations.png
       :alt: Flow chart final result exit point calculations
       :align: center

    :param geoprob_pipe:
    :param results:
    :return:
    """

    df_beta_scenarios_final = results.df_beta_scenarios_final.copy(deep=True)

    # Sum
    df = df_beta_scenarios_final.assign(
        failure_probability=df_beta_scenarios_final.apply(
            lambda row: row['failure_probability'] * geoprob_pipe.input_data.scenarios.scenario_kans(
                vak_id=row['vak_id'], scenario_naam=row['ondergrondscenario_id']
            ), axis=1)).groupby('uittredepunt_id', as_index=False)[
        'failure_probability'].sum()
    df["beta"] = df["failure_probability"].apply(lambda failure_prob: convert_failure_probability_to_beta(failure_prob))

    # Determine when uittredepunt is converged (when all scenarios are converged)
    conv = df_beta_scenarios_final.groupby('uittredepunt_id', as_index=False)["converged"].all()
    df = df.merge(conv, on="uittredepunt_id", how="left")

    # Determine uittredepunt flow_chart_number
    flow_chart_number = df_beta_scenarios_final.groupby('uittredepunt_id', as_index=False)["flow_chart_number"].max()
    df = df.merge(flow_chart_number, on="uittredepunt_id", how="left")
    df['advise'] = df['flow_chart_number'].map({5: "Consider fine tuning on scenario-level."}).fillna("-")
    df['flow_chart_number'] = df['flow_chart_number'].map({5: 11}).fillna(12)

    # Add vak id back to it
    gdf_uittredepunten = geoprob_pipe.input_data.uittredepunten.gdf
    df_uittredepunten = gdf_uittredepunten[["uittredepunt_id", "vak_id"]]
    df = df.merge(df_uittredepunten, left_on="uittredepunt_id", right_on="uittredepunt_id")

    return df[["uittredepunt_id", "vak_id", "converged", "beta", "failure_probability", "advise", "flow_chart_number"]]


def construct_df_beta_per_vak(results: Results) -> DataFrame:
    """ Constructs the DataFrame of the final result for the vakken.

    Because there is an automated decision-making in the scenario and exit point calculations (see flow charts over
    there), for the vakken the flow chart is extended below.

    .. image:: /_static/flow-chart-final-result-vak-calculations.png
       :alt: Flow chart final result vak calculations
       :align: center

    :param results:
    :return:
    """

    # Gather data
    df_beta_uittredepunten = results.df_beta_uittredepunten.copy(deep=True)

    # Minimale beta van beta uittredepunten per vak
    df: DataFrame = df_beta_uittredepunten.loc[df_beta_uittredepunten.groupby('vak_id')['beta'].idxmin()]
    df = df.drop(columns=["flow_chart_number"])

    # Determine vak flow chart number
    flow_chart_number = df_beta_uittredepunten.groupby('uittredepunt_id', as_index=False)["flow_chart_number"].min()
    df = df.merge(flow_chart_number, on="uittredepunt_id", how="left")
    df['advise'] = df['flow_chart_number'].map({11: "Consider fine tuning on scenario-level."}).fillna("-")
    df['flow_chart_number'] = df['flow_chart_number'].map({11: 21}).fillna(22)

    return df[["vak_id", "beta", "failure_probability", "advise"]]



"""
TODO Nu Must Klein: Voeg laag toe aan GeoPackage met visuele koppeling tussen HRD en uittredepunten.
"""

from __future__ import annotations
from geopandas import GeoDataFrame, read_file
from pathlib import Path
from pandas import DataFrame
from geoprob_pipe.cmd_app.spatial_joins.utils import append_to_gis_join_parameter_invoer_table
from geoprob_pipe.utils.validation_messages import BColors
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def coupled_polderpeil_to_uittredepunten(app_settings: ApplicationSettings) -> bool:


    # Getting necessary GeoDataframes
    gdf_exit_points: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="uittredepunten")
    gdf_polderpeil: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="polderpeil")

    # Check if already added
    if "polderpeil" in gdf_exit_points.columns:
        print(BColors.OKBLUE, f"✔  Polderpeil al gekoppeld aan uittredepunten.", BColors.ENDC)
        return True  # Assuming already added

    # Perform spatial join to find the nearest HRD-location for each Exit Point
    gdf_exit_with_hrd = gdf_exit_points.sjoin_nearest(
        gdf_polderpeil[['geometry', 'polderpeil']], how='left', distance_col='distance')

    # Define which columns to keep (after spatial join)
    columns_to_keep = list(gdf_exit_points.columns)
    columns_to_keep.append("polderpeil")
    gdf_new_exit_points = gdf_exit_with_hrd[columns_to_keep]

    # Append to new version of geospatial storage
    df_to_append: DataFrame = gdf_new_exit_points[["polderpeil", "uittredepunt_id"]]
    df_to_append = df_to_append.rename(columns={"polderpeil": "mean"})
    append_to_gis_join_parameter_invoer_table(
        df_sjoin=df_to_append,
        parameter_name="polderpeil",
        geopackage_filepath=app_settings.geopackage_filepath)

    # Store back in geopackage
    gdf_new_exit_points.to_file(Path(app_settings.geopackage_filepath), layer="uittredepunten", driver="GPKG")
    # TODO: Uitfaseren deze oude manier van polderpeil opslaan
    print(BColors.OKBLUE, f"✅  Polderpeil is nu gekoppeld aan de uittredepunten.", BColors.ENDC)
    return True



from __future__ import annotations
import os
import plotly.graph_objects as go
import pandas as pd
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.comparisons import ComparisonCollector


def _add_traces(comparison: ComparisonCollector, df: pd.DataFrame, fig: go.Figure):

    hoverdata = df[["uittredepunt_id", "beta1", "beta2"]].to_numpy()
    symbol_map = {
        1: "circle",
        0: "x"
    }

    symbols1 = [symbol_map[b] for b in df["converged1"]]
    symbols2 = [symbol_map[b] for b in df["converged2"]]
    name1 = "Beta1: " + comparison.name_1 + ".geoprob_pipe.gpkg"
    name2 = "Beta2: " + comparison.name_2 + ".geoprob_pipe.gpkg"

    for _, row in df.iterrows():
        fig.add_trace(go.Scatter(
            x=[row["uittredepunt_id"], row["uittredepunt_id"]],
            y=[row["beta1"], row["beta2"]],
            mode="lines",
            showlegend=False,
            marker=dict(
                color="grey"
            )
        ))

    fig.add_trace(go.Scatter(
        x=df["uittredepunt_id"],
        y=df["beta1"],
        mode="markers",
        name=name1,
        marker=dict(
            symbol=symbols1,
            color="green",
            size=10
            ),
        customdata=hoverdata,
        hovertemplate=(
            "Uittredepunt ID: %{customdata[0]}<br>"
            "β1: %{customdata[1]:.2f}<br>"
            "β2: %{customdata[2]:.2f}<br>"
            "<extra></extra>"
        ),
        legendgroup="beta1",
        showlegend=False
    ))

    fig.add_trace(go.Scatter(
        x=df["uittredepunt_id"], y=df["beta2"], mode="markers",
        name=name2,
        marker=dict(symbol=symbols2, color="blue", size=10),
        customdata=hoverdata,
        hovertemplate=(
            "Uittredepunt ID: %{customdata[0]}<br>"
            "β2: %{customdata[2]:.2f}<br>"
            "β1: %{customdata[1]:.2f}<br>"
            "<extra></extra>"
        ),
        legendgroup="beta2",
        showlegend=False
    ))
    # Empty trace for legend
    fig.add_trace(go.Scatter(
        x=[None], y=[None], mode="markers",
        marker=dict(symbol="circle", color="green", size=10),
        name=name1,
        legendgroup="beta1",
    ))
    fig.add_trace(go.Scatter(
        x=[None], y=[None], mode="markers",
        marker=dict(symbol="circle", color="blue", size=10),
        name=name2,
        legendgroup="beta2"
    ))
    fig.add_trace(go.Scatter(
        x=[None], y=[None], mode="markers", name="Not converged",
        marker=dict(symbol="x", color="white", size=10, line=dict(color="black", width=0.5))))
    return fig


def _add_vak_id(comparison: ComparisonCollector, fig: go.Figure):

    vak_ids = comparison.df1_beta_uittredepunten["vak_id"].unique()
    for _, vak_id in enumerate(vak_ids):
        df_vak = comparison.df1_beta_uittredepunten
        uit_ids = df_vak.loc[df_vak["vak_id"] == vak_id, ["uittredepunt_id"]]

        min_id: int = uit_ids.min().iloc[0]
        max_id: int = uit_ids.max().iloc[0]

        fig.add_trace(go.Scatter(
            x=[min_id, max_id], y=[0, 0], mode="lines",
            line=dict(color="red", width=2), showlegend=False, name=f"{vak_id}"))

        fig.add_annotation(
            x=(min_id + max_id) / 2, y=-0.7, text=f"{vak_id}", showarrow=False,
            font=dict(size=10, color="black"), align="center")
    return fig


def dumbbell_beta(comparison: ComparisonCollector,
                  export: bool = False):

    df_result1 = (comparison.df1_beta_uittredepunten[
        ["uittredepunt_id", "converged", "beta"]
        ].rename(columns={"beta": "beta1", "converged": "converged1"}))
    df_result2 = (comparison.df2_beta_uittredepunten[
        ["uittredepunt_id", "converged", "beta"]
        ].rename(columns={"beta": "beta2", "converged": "converged2"}))

    df = df_result1.merge(df_result2, on="uittredepunt_id")

    fig = go.Figure()
    fig = _add_traces(comparison, df, fig)
    fig = _add_vak_id(comparison, fig)

    fig.update_layout(
        title=f"Vergelijk Beta tussen GeoProb-Pipe bestanden<br>"
              f"<sup>Voor de <b>gecombineerde limit state</b>, en gevisualiseerd per uittredepunt.</sup>",
        xaxis_title="Uittredepunt ID",
        yaxis_title="β-value",
        legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1))

    if export:
        os.makedirs(comparison.export_dir, exist_ok=True)
        fig.write_html(os.path.join(comparison.export_dir, "dumbbell_beta.html"), include_plotlyjs='cdn')

    return fig


def dumbbell_uplift(comparison: ComparisonCollector,
                    export: bool = False):

    df_result1 = (comparison.df1_beta_limit_states[
        ["uittredepunt_id", "limit_state", "beta", "ondergrondscenario_id",
         "converged"]].rename(columns={"beta": "beta1",
                                       "limit_state": "limit_state1",
                                       "converged": "converged1"}))
    df_result1 = df_result1[df_result1["limit_state1"] == "calc_Z_u"]

    df_result2 = (comparison.df2_beta_limit_states[
        ["uittredepunt_id", "limit_state", "beta", "ondergrondscenario_id",
         "converged"]].rename(columns={"beta": "beta2",
                                       "limit_state": "limit_state2",
                                       "converged": "converged2"}))
    df_result2 = df_result2[df_result2["limit_state2"] == "calc_Z_u"]

    df = df_result1.merge(df_result2, on=["uittredepunt_id",
                                          "ondergrondscenario_id"])
    fig_list = []
    for scenario in df["ondergrondscenario_id"].unique():
        fig = go.Figure()
        scenario_mask = df["ondergrondscenario_id"] == scenario
        df_plot = df[scenario_mask]
        fig = _add_traces(comparison, df_plot, fig)
        fig = _add_vak_id(comparison, fig)

        fig.update_layout(
            title=f"Vergelijk Beta tussen GeoProb-Pipe bestanden<br>"
                  f"<sup>Voor de <b>limit state Uplift</b> en <b>scenario {scenario}</b>, "
                  f"en gevisualiseerd per uittredepunt.</sup>",
            xaxis_title="Uittredepunt ID",
            yaxis_title="β-value",
            legend=dict(
                orientation="h",
                yanchor="bottom",
                y=1.02,
                xanchor="right",
                x=1
            )
        )
        fig_list.append(fig)
        if export:
            os.makedirs(comparison.export_dir, exist_ok=True)
            fig.write_html(os.path.join(
                comparison.export_dir, f"dumbbell_uplift_{scenario}.html"), include_plotlyjs='cdn')
            # fig.write_image(os.path.join(
            #     comparison.export_dir, f"dumbbell_uplift_{scenario}.png"
            #     ), format="png", scale=5,  width=1400)
    return fig_list


def dumbbell_heave(comparison: ComparisonCollector,
                   export: bool = False):
    df_result1 = (comparison.df1_beta_limit_states[
        ["uittredepunt_id", "limit_state", "beta", "ondergrondscenario_id",
         "converged"]].rename(columns={"beta": "beta1",
                                       "limit_state": "limit_state1",
                                       "converged": "converged1"}))
    df_result1 = df_result1[df_result1["limit_state1"] == "calc_Z_h"]
    df_result2 = (comparison.df2_beta_limit_states[
        ["uittredepunt_id", "limit_state", "beta", "ondergrondscenario_id",
         "converged"]].rename(columns={"beta": "beta2",
                                       "limit_state": "limit_state2",
                                       "converged": "converged2"}))
    df_result2 = df_result2[df_result2["limit_state2"] == "calc_Z_h"]

    df = df_result1.merge(df_result2, on=["uittredepunt_id",
                                          "ondergrondscenario_id"])
    fig_list = []
    for scenario in df["ondergrondscenario_id"].unique():
        fig = go.Figure()
        scenario_mask = df["ondergrondscenario_id"] == scenario
        df_plot = df[scenario_mask]
        fig = _add_traces(comparison, df_plot, fig)
        fig = _add_vak_id(comparison, fig)

        fig.update_layout(
            title=f"Vergelijk Beta tussen GeoProb-Pipe bestanden<br>"
                  f"<sup>Voor de <b>limit state Heave</b> en <b>scenario {scenario}</b>, "
                  f"en gevisualiseerd per uittredepunt.</sup>",
            xaxis_title="Uittredepunt ID",
            yaxis_title="β-value",
            legend=dict(
                orientation="h",
                yanchor="bottom",
                y=1.02,
                xanchor="right",
                x=1
            )
        )
        fig_list.append(fig)
        if export:
            os.makedirs(comparison.export_dir, exist_ok=True)
            fig.write_html(os.path.join(
                comparison.export_dir, f"dumbbell_heave_{scenario}.html"
                ), include_plotlyjs='cdn')
            # fig.write_image(os.path.join(
            #     comparison.export_dir, f"dumbbell_heave_{scenario}.png"
            #     ), format="png", scale=5, width=1400)
    return fig_list


def dumbbell_piping(comparison: ComparisonCollector,
                    export: bool = False):
    df_result1 = (comparison.df1_beta_limit_states[
        ["uittredepunt_id", "limit_state", "beta", "ondergrondscenario_id",
         "converged"]].rename(columns={
             "beta": "beta1",
             "limit_state": "limit_state1",
             "converged": "converged1"}))
    df_result1 = df_result1[df_result1["limit_state1"] == "calc_Z_p"]
    df_result2 = (comparison.df2_beta_limit_states[
        ["uittredepunt_id", "limit_state", "beta", "ondergrondscenario_id",
         "converged"]].rename(columns={
             "beta": "beta2",
             "limit_state": "limit_state2",
             "converged": "converged2"}))
    df_result2 = df_result2[df_result2["limit_state2"] == "calc_Z_p"]

    df = df_result1.merge(df_result2, on=["uittredepunt_id",
                                          "ondergrondscenario_id"])
    fig_list = []
    for scenario in df["ondergrondscenario_id"].unique():
        fig = go.Figure()
        scenario_mask = df["ondergrondscenario_id"] == scenario
        df_plot = df[scenario_mask]
        fig = _add_traces(comparison, df_plot, fig)
        fig = _add_vak_id(comparison, fig)

        fig.update_layout(
            title=f"Vergelijk Beta tussen GeoProb-Pipe bestanden<br>"
                  f"<sup>Voor de <b>limit state Piping</b> en <b>scenario {scenario}</b>, "
                  f"en gevisualiseerd per uittredepunt.</sup>",
            xaxis_title="Uittredepunt ID",
            yaxis_title="β-value",
            legend=dict(
                orientation="h",
                yanchor="bottom",
                y=1.02,
                xanchor="right",
                x=1
            )
        )
        fig_list.append(fig)
        if export:
            os.makedirs(comparison.export_dir, exist_ok=True)
            fig.write_html(os.path.join(
                comparison.export_dir, f"dumbbell_piping_{scenario}.html"
                ), include_plotlyjs='cdn')
            # fig.write_image(os.path.join(
            #     comparison.export_dir, f"dumbbell_piping_{scenario}.png"
            #     ), format="png", scale=5, width=1400)
    return fig_list



from __future__ import annotations
from geoprob_pipe.visualizations.graphs.betrouwbaarheidsindex import ( 
    GraphBetaValuesSingleInteractive)
    # beta_uittredepunten_graph, beta_scenarios_graph, beta_vakken_graph)
from geoprob_pipe.visualizations.graphs.hfreq import GraphHFreqSingleInteractive
from geoprob_pipe.visualizations.graphs.physical_values_along_levee import physical_values_buitenwaterstand_and_top_zand
from geoprob_pipe.visualizations.graphs.invloedsfactoren import invloedsfactoren
from geoprob_pipe.visualizations.graphs.phreatic_waterline import phreatic_waterline
from geoprob_pipe.visualizations.graphs.overview_alpha import overview_alpha
from geoprob_pipe.visualizations.graphs.river_waterlevel import river_waterlevel
from typing import TYPE_CHECKING
from plotly.graph_objects import Figure as PlotlyFigure
import os

if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


class Graphs:

    def __init__(self, app_obj: GeoProbPipe):
        self.geoprob_pipe = app_obj

    @property
    def export_dir(self) -> str:
        path = os.path.join(self.geoprob_pipe.visualizations.export_dir, "graphs")
        os.makedirs(path, exist_ok=True)
        return path

    def hfreq_graph_in_single_interactive(self, export: bool = False) -> PlotlyFigure:
        graph = GraphHFreqSingleInteractive(self.geoprob_pipe, export=export)
        return graph.fig

    def invloedsfactoren(self, export: bool = False) -> PlotlyFigure:
        return invloedsfactoren(self.geoprob_pipe, export=export)

    def physical_values_buitenwaterstand_and_top_zand(self, export: bool = False) -> PlotlyFigure:
        return physical_values_buitenwaterstand_and_top_zand(self.geoprob_pipe, export=export)

    def beta_value_in_single_interactive(self, export: bool = False) -> PlotlyFigure:
        graph = GraphBetaValuesSingleInteractive(self.geoprob_pipe, export=export)
        return graph.fig

    # def beta_scenarios(self) -> PlotlyFigure:
    #     return beta_scenarios_graph(self.geoprob_pipe, export=False)
    #
    # def beta_uittredepunten(self) -> PlotlyFigure:
    #     return beta_uittredepunten_graph(self.geoprob_pipe, export=False)
    #
    # def beta_vakken(self) -> PlotlyFigure:
    #     return beta_vakken_graph(self.geoprob_pipe, export=False)

    def phreatic_waterline(self) -> PlotlyFigure:
        return phreatic_waterline(self.geoprob_pipe, export=False)

    def overview_alpha(self) -> PlotlyFigure:
        return overview_alpha(self.geoprob_pipe, export=False)

    def river_waterlevel(self) -> PlotlyFigure:
        return river_waterlevel(self.geoprob_pipe, export=False)

    def export_graphs(self):
        GraphHFreqSingleInteractive(self.geoprob_pipe, export=True)
        GraphBetaValuesSingleInteractive(self.geoprob_pipe, export=True)
        # beta_scenarios_graph(self.geoprob_pipe, export=True)
        # beta_uittredepunten_graph(self.geoprob_pipe, export=True)
        # beta_vakken_graph(self.geoprob_pipe, export=True)
        self.physical_values_buitenwaterstand_and_top_zand(export=True)
        self.invloedsfactoren(export=True)
        phreatic_waterline(self.geoprob_pipe, export=True)
        overview_alpha(self.geoprob_pipe, export=True)
        river_waterlevel(self.geoprob_pipe, export=True)
